from typing import Optional, Tuple, Union

import pyro
import torch
from torch.distributions import (
    Bernoulli,
    NegativeBinomial,
)
from packaging import version
from pyro import poutine
from pyro.distributions import (
    Bernoulli,  # shadowing from torch.distributions for pyright
    Beta,
    LogNormal,
    NegativeBinomial,  # shadowing from torch.distributions for pyright
    Normal,
    Poisson,
)
from pyro.distributions.constraints import positive
from pyro.nn import PyroModule, PyroParam, PyroSample
from pyro.primitives import plate
from scvi.nn import Decoder
from torch.nn.functional import relu, softplus

from pyrovelocity.utils import mRNA


class LogNormalModel(PyroModule):
    """
    A base class for pyrovelocity models.

    This class serves as the base class for constructing a pyrovelocity model.
    It provides basic methods for handling gene expression data, such as creating plates
    for cells and genes, encoding cell-specific features, and computing the likelihood
    of the observed data.

    Attributes:
        num_cells (int): The number of cells.
        num_genes (int): The number of genes.
        likelihood (str): The likelihood type for the model, defaults to "Poisson".
        plate_size (int): The size of the plate for the model, defaults to 2.

    Example:
        >>> from pyrovelocity._velocity_model_deprecated import LogNormalModel
        >>> num_cells = 10
        >>> num_genes = 20
        >>> likelihood = "Poisson"
        >>> plate_size = 2
        >>> model = LogNormalModel(num_cells, num_genes, likelihood, plate_size)
        >>> assert model.num_cells == num_cells
        >>> assert model.num_genes == num_genes
        >>> assert model.likelihood == likelihood
        >>> assert model.plate_size == plate_size
    """

    def __init__(
        self,
        num_cells: int,
        num_genes: int,
        likelihood: str = "Poisson",
        plate_size: int = 2,
        correct_library_size: Union[bool, str] = True,
    ) -> None:
        assert num_cells > 0 and num_genes > 0
        super().__init__()
        self.num_cells = num_cells
        self.num_genes = num_genes
        self.n_obs = None
        self.plate_size = plate_size
        self.correct_library_size = correct_library_size
        self.register_buffer("zero", torch.tensor(0.0))
        self.register_buffer("one", torch.tensor(1.0))

        self.likelihood = likelihood
        if self.likelihood == "NB":
            self.u_px_r = PyroParam(
                torch.ones(self.num_genes), constraint=positive, event_dim=0
            )
            self.s_px_r = PyroParam(
                torch.ones(self.num_genes), constraint=positive, event_dim=0
            )

    def create_plates(
        self,
        u_obs: Optional[torch.Tensor] = None,
        s_obs: Optional[torch.Tensor] = None,
        u_log_library: Optional[torch.Tensor] = None,
        s_log_library: Optional[torch.Tensor] = None,
        u_log_library_loc: Optional[torch.Tensor] = None,
        s_log_library_loc: Optional[torch.Tensor] = None,
        u_log_library_scale: Optional[torch.Tensor] = None,
        s_log_library_scale: Optional[torch.Tensor] = None,
        ind_x: Optional[torch.Tensor] = None,
        cell_state: Optional[torch.Tensor] = None,
        time_info: Optional[torch.Tensor] = None,
    ) -> Tuple[plate, plate]:
        cell_plate = pyro.plate(
            "cells", self.num_cells, subsample=ind_x, dim=-2
        )
        gene_plate = pyro.plate("genes", self.num_genes, dim=-1)
        return cell_plate, gene_plate

    def create_plate(
        self,
        u_obs: Optional[torch.Tensor] = None,
        s_obs: Optional[torch.Tensor] = None,
        u_log_library: Optional[torch.Tensor] = None,
        s_log_library: Optional[torch.Tensor] = None,
        ind_x: Optional[torch.Tensor] = None,
    ):
        return pyro.plate("cells", self.num_cells, subsample=ind_x, dim=-1)

    @PyroSample
    def alpha(self):
        if self.plate_size == 2:
            return LogNormal(self.zero, self.one)
        return (
            LogNormal(self.zero, self.one * 0.1)
            .expand((self.num_genes,))
            .to_event(1)
            .mask(False)
        )

    @PyroSample
    def t_scale(self):
        return Normal(self.zero, self.one * 0.1)

    @PyroSample
    def beta(self):
        if self.plate_size == 2:
            return LogNormal(self.zero, self.one * 0.25)
        return (
            LogNormal(self.zero, self.one * 0.1)
            .expand((self.num_genes,))
            .to_event(1)
            .mask(False)
        )

    @PyroSample
    def gamma(self):
        if self.plate_size == 2:
            return LogNormal(self.zero, self.one)
        return (
            LogNormal(self.zero, self.one * 0.1)
            .expand((self.num_genes,))
            .to_event(1)
            .mask(False)
        )

    @PyroSample
    def gene_offset(self):
        return Normal(self.zero, self.one)

    @PyroSample
    def u_scale(self):
        if self.plate_size == 2:
            return LogNormal(self.zero, self.one * 0.1)
        return (
            LogNormal(self.zero, self.one * 0.1)
            .expand((self.num_genes,))
            .to_event(1)
            .mask(False)
        )

    @PyroSample
    def s_scale(self):
        if self.plate_size == 2:
            return LogNormal(self.zero, self.one * 0.1)
        return (
            LogNormal(self.zero, self.one * 0.1)
            .expand((self.num_genes,))
            .to_event(1)
            .mask(False)
        )

    @PyroSample
    def u_inf(self):
        if self.plate_size == 2:
            return LogNormal(self.zero, self.one * 0.1)
        return (
            LogNormal(self.zero, self.one * 0.1)
            .expand((self.num_genes,))
            .to_event(1)
            .mask(False)
        )

    @PyroSample
    def s_inf(self):
        if self.plate_size == 2:
            return LogNormal(self.zero, self.one * 0.1)
        return (
            LogNormal(self.zero, self.one * 0.1)
            .expand((self.num_genes,))
            .to_event(1)
            .mask(False)
        )

    @PyroSample
    def dt_switching(self):
        if self.plate_size == 2:
            return LogNormal(self.zero, self.one)
        return (
            LogNormal(self.zero, self.one * 0.1)
            .expand((self.num_genes,))
            .to_event(1)
            .mask(False)
        )

    @PyroSample
    def latent_time(self):
        if self.shared_time:
            if self.plate_size == 2:
                return Normal(self.zero, self.one * 0.1).mask(
                    self.include_prior
                )
            else:
                return (
                    Normal(self.zero, self.one * 0.1)
                    .expand((self.num_genes,))
                    .to_event(1)
                    .mask(self.include_prior)
                )
        if self.plate_size == 2:
            return LogNormal(self.zero, self.one).mask(self.include_prior)
        return (
            LogNormal(self.zero, self.one).expand((self.num_genes,)).to_event(1)
        )

    @PyroSample
    def cell_time(self):
        if self.plate_size == 2 and self.shared_time:
            return Normal(self.zero, self.one)
        else:
            return LogNormal(self.zero, self.one)

    @PyroSample
    def p_velocity(self):
        return Beta(self.one, self.one)

    @PyroSample
    def cell_codebook(self):
        return (
            Normal(self.zero, self.one / (10 * self.latent_factor_size))
            .expand((self.latent_factor_size, self.num_genes * 2))
            .to_event(2)
            .mask(self.include_prior)
        )

    @PyroSample
    def cell_code(self):
        return (
            Normal(self.zero, self.one)
            .expand((self.latent_factor_size,))
            .to_event(1)
            .mask(self.include_prior)
        )

    def get_likelihood(
        self,
        ut: torch.Tensor,
        st: torch.Tensor,
        u_log_library: Optional[torch.Tensor] = None,
        s_log_library: Optional[torch.Tensor] = None,
        u_scale: Optional[torch.Tensor] = None,
        s_scale: Optional[torch.Tensor] = None,
        u_read_depth: Optional[torch.Tensor] = None,
        s_read_depth: Optional[torch.Tensor] = None,
        u_cell_size_coef: None = None,
        ut_coef: None = None,
        s_cell_size_coef: None = None,
        st_coef: None = None,
    ) -> Tuple[Poisson, Poisson]:
        """
        Compute the likelihood of the given count data.

        Args:
            ut (torch.Tensor): Tensor representing unspliced transcripts.
            st (torch.Tensor): Tensor representing spliced transcripts.
            u_log_library (Optional[torch.Tensor], optional): Log library tensor for unspliced transcripts. Defaults to None.
            s_log_library (Optional[torch.Tensor], optional): Log library tensor for spliced transcripts. Defaults to None.
            u_scale (Optional[torch.Tensor], optional): Scale tensor for unspliced transcripts. Defaults to None.
            s_scale (Optional[torch.Tensor], optional): Scale tensor for spliced transcripts. Defaults to None.
            u_read_depth (Optional[torch.Tensor], optional): Read depth tensor for unspliced transcripts. Defaults to None.
            s_read_depth (Optional[torch.Tensor], optional): Read depth tensor for spliced transcripts. Defaults to None.
            u_cell_size_coef (Optional[Any], optional): Cell size coefficient for unspliced transcripts. Defaults to None.
            ut_coef (Optional[Any], optional): Coefficient for unspliced transcripts. Defaults to None.
            s_cell_size_coef (Optional[Any], optional): Cell size coefficient for spliced transcripts. Defaults to None.
            st_coef (Optional[Any], optional): Coefficient for spliced transcripts. Defaults to None.

        Returns:
            Tuple[Poisson, Poisson]: A tuple of Poisson distributions for unspliced and spliced transcripts, respectively.

        Example:
            >>> import torch
            >>> from pyro.nn import PyroModule
            >>> num_cells = 10
            >>> num_genes = 20
            >>> likelihood = "Poisson"
            >>> plate_size = 2
            >>> model = LogNormalModel(num_cells, num_genes, likelihood, plate_size)
            >>> ut = torch.rand(num_cells, num_genes)
            >>> st = torch.rand(num_cells, num_genes)
            >>> u_read_depth = torch.rand(num_cells, 1)
            >>> s_read_depth = torch.rand(num_cells, 1)
            >>> u_dist, s_dist = model.get_likelihood(ut, st, u_read_depth=u_read_depth, s_read_depth=s_read_depth)
            >>> assert isinstance(u_dist, torch.distributions.Poisson)
            >>> assert isinstance(s_dist, torch.distributions.Poisson)
        """
        if self.likelihood == "NB":
            if self.correct_library_size:
                ut = relu(ut) + self.one * 1e-6
                st = relu(st) + self.one * 1e-6
                ut = pyro.deterministic("ut", ut, event_dim=0)
                st = pyro.deterministic("st", st, event_dim=0)
                if self.guide_type not in [
                    "velocity_auto",
                    "velocity_auto_depth",
                ]:
                    ut = ut / torch.sum(ut, dim=-1, keepdim=True)
                    st = st / torch.sum(st, dim=-1, keepdim=True)
                ut = pyro.deterministic("ut_norm", ut, event_dim=0)
                st = pyro.deterministic("st_norm", st, event_dim=0)
                u_logits = ((ut + self.one * 1e-6) * u_read_depth).log() - (
                    self.u_px_r.exp() + self.one * 1e-6
                ).log()
                s_logits = ((st + self.one * 1e-6) * s_read_depth).log() - (
                    self.s_px_r.exp() + self.one * 1e-6
                ).log()
            else:
                ut = relu(ut)
                st = relu(st)
                ut = pyro.deterministic("ut", ut, event_dim=0)
                st = pyro.deterministic("st", st, event_dim=0)
                u_logits = (relu(ut) + self.one * 1e-6).log() - (
                    self.u_px_r.exp() + self.one * 1e-6
                ).log()
                s_logits = (relu(st) + self.one * 1e-6).log() - (
                    self.s_px_r.exp() + self.one * 1e-6
                ).log()
            u_dist = NegativeBinomial(
                total_count=self.u_px_r.exp(), logits=u_logits
            )
            s_dist = NegativeBinomial(
                total_count=self.s_px_r.exp(), logits=s_logits
            )
        elif self.likelihood == "Poisson":
            if self.correct_library_size:
                ut = relu(ut) + self.one * 1e-6
                st = relu(st) + self.one * 1e-6
                ut = pyro.deterministic("ut", ut, event_dim=0)
                st = pyro.deterministic("st", st, event_dim=0)
                if self.correct_library_size == "cell_size_regress":
                    ut_sum = torch.log(torch.sum(ut, dim=-1, keepdim=True))
                    st_sum = torch.log(torch.sum(st, dim=-1, keepdim=True))
                    ut = torch.log(ut)
                    st = torch.log(st)
                    ut = torch.exp(
                        ut_coef * ut
                        + u_cell_size_coef * (-ut_sum + u_read_depth)
                    )
                    st = torch.exp(
                        st_coef * st
                        + s_cell_size_coef * (-st_sum + s_read_depth)
                    )
                else:
                    ut = ut / torch.sum(ut, dim=-1, keepdim=True)
                    st = st / torch.sum(st, dim=-1, keepdim=True)
                ut = pyro.deterministic("ut_norm", ut, event_dim=0)
                st = pyro.deterministic("st_norm", st, event_dim=0)

                if self.correct_library_size != "cell_size_regress":
                    ut = (ut + self.one * 1e-6) * u_read_depth
                    st = (st + self.one * 1e-6) * s_read_depth
            else:
                ut = relu(ut)
                st = relu(st)
                ut = pyro.deterministic("ut", ut, event_dim=0)
                st = pyro.deterministic("st", st, event_dim=0)
                ut = ut + self.one * 1e-6
                st = st + self.one * 1e-6

            u_dist = Poisson(ut)
            s_dist = Poisson(st)
        elif self.likelihood == "Normal":
            if u_scale is not None and s_scale is not None:
                u_dist = Normal(ut, u_scale)
                s_dist = Normal(st, s_scale)
            else:
                u_dist = Normal(ut, self.one * 0.1)
                s_dist = Normal(st, self.one * 0.1)
        elif self.likelihood == "LogNormal":
            if u_scale is not None and s_scale is not None:
                u_dist = LogNormal((ut + self.one * 1e-6).log(), u_scale)
                s_dist = LogNormal((st + self.one * 1e-6).log(), s_scale)
            else:
                u_dist = LogNormal(ut, self.one * 0.1)
                s_dist = LogNormal(st, self.one * 0.1)
        else:
            raise
        return u_dist, s_dist


class VelocityModel(LogNormalModel):
    def __init__(
        self,
        num_cells: int,
        num_genes: int,
        likelihood: str = "Poisson",
        shared_time: bool = True,
        t_scale_on: bool = False,
        plate_size: int = 2,
        latent_factor: str = "none",
        latent_factor_size: int = 30,
        latent_factor_operation: str = "selection",
        include_prior: bool = False,
        num_aux_cells: int = 100,
        only_cell_times: bool = False,
        decoder_on: bool = False,
        add_offset: bool = False,
        correct_library_size: Union[bool, str] = True,
        guide_type: bool = "velocity",
        cell_specific_kinetics: Optional[str] = None,
        kinetics_num: Optional[int] = None,
        **initial_values,
    ) -> None:
        """
        Derived class for modeling RNA velocity with shared latent time.

        Inherits from LogNormalModel.

        Args:
            num_cells (int): Number of cells.
            num_genes (int): Number of genes.
            likelihood (str, optional): Likelihood model. Defaults to "Poisson".
            shared_time (bool, optional): Whether to share time across cells. Defaults to True.
            t_scale_on (bool, optional): Whether to scale the time. Defaults to False.
            plate_size (int, optional): The number of dimensions for the plates. Defaults to 2.
            latent_factor (str, optional): Latent factor type. Defaults to "none".
            latent_factor_size (int, optional): Latent factor size. Defaults to 30.
            latent_factor_operation (str, optional): Latent factor operation. Defaults to "selection".
            include_prior (bool, optional): Whether to include prior. Defaults to False.
            num_aux_cells (int, optional): Number of auxiliary cells. Defaults to 100.
            only_cell_times (bool, optional): Whether to use only cell times. Defaults to False.
            decoder_on (bool, optional): Whether to use a decoder. Defaults to False.
            add_offset (bool, optional): Whether to add an offset. Defaults to False.
            correct_library_size (Union[bool, str], optional): Whether to correct library size. Defaults to True.
            guide_type (bool, optional): Guide type for the model. Defaults to "velocity".
            cell_specific_kinetics (Optional[str], optional): Cell-specific kinetics. Defaults to None.
            kinetics_num (Optional[int], optional): Number of kinetics. Defaults to None.
            **initial_values: Initial values for various parameters in the model.

        Example:
            >>> import torch
            >>> from pyro.nn import PyroModule
            >>> num_cells = 10
            >>> num_genes = 20
            >>> likelihood = "Poisson"
            >>> plate_size = 2
            >>> model = VelocityModel(num_cells, num_genes, likelihood, plate_size)
            >>> u_obs = torch.rand(num_cells, num_genes)
            >>> s_obs = torch.rand(num_cells, num_genes)
            >>> u_log_library = torch.rand(num_cells, 1)
            >>> s_log_library = torch.rand(num_cells, 1)
            >>> ind_x = torch.randint(0, num_cells, (num_cells,))
            >>> # ut, st = model.model(u_obs, s_obs, u_log_library, s_log_library, ind_x)
            >>> # assert ut.shape == (num_cells, num_genes)
            >>> # assert st.shape == (num_cells, num_genes)
        """
        assert num_cells > 0 and num_genes > 0
        super().__init__(num_cells, num_genes, likelihood, plate_size)
        self.num_aux_cells = num_aux_cells
        self.num_aux_cells = num_aux_cells
        self.only_cell_times = only_cell_times
        self.guide_type = guide_type
        self.cell_specific_kinetics = cell_specific_kinetics
        self.k = kinetics_num

        self.mask = initial_values.get(
            "mask", torch.ones(self.num_cells, self.num_genes).bool()
        )
        for key in initial_values:
            self.register_buffer(f"{key}_init", initial_values[key])

        self.shared_time = shared_time
        self.t_scale_on = t_scale_on
        self.add_offset = add_offset
        self.plate_size = plate_size

        self.latent_factor = latent_factor
        self.latent_factor_size = latent_factor_size
        self.latent_factor_operation = latent_factor_operation
        self.include_prior = include_prior
        self.decoder_on = decoder_on
        self.correct_library_size = correct_library_size
        if self.decoder_on:
            self.decoder = Decoder(1, self.num_genes, n_layers=2)


class AuxCellVelocityModel(VelocityModel):
    def generate_cell(
        self,
        gene_plate,
        alpha,
        beta,
        gamma,
        switching,
        u_inf,
        s_inf,
        u_scale,
        s_scale,
        u_log_library,
        s_log_library,
        u_pcs_mean=None,
        s_pcs_mean=None,
        cell_codebook=None,
        t_scale=None,
        gene_offset=None,
        p_velocity=None,
        decoder_on=False,
        u_read_depth=None,
        s_read_depth=None,
        u0=None,
        s0=None,
    ):
        if self.decoder_on:
            pyro.module("time_decoder", self.decoder)
        if self.latent_factor == "linear":
            cell_code = pyro.sample(
                "cell_code",
                Normal(self.zero, self.one)
                .expand((self.latent_factor_size,))
                .to_event(1)
                .mask(self.include_prior),
            )
        else:
            cell_code = None

        if self.shared_time:
            cell_time = pyro.sample(
                "cell_time",
                LogNormal(self.zero, self.one).mask(self.include_prior),
            )

        with gene_plate:
            if self.latent_factor_operation == "selection":
                cellgene_type = pyro.sample(
                    "cellgene_type", Bernoulli(p_velocity)
                )
            else:
                cellgene_type = None

            if self.only_cell_times:
                if self.decoder_on:
                    t, _ = self.decoder(cell_time)
                else:
                    t = cell_time

            elif self.shared_time:
                t = pyro.sample(
                    "latent_time",
                    Normal(self.zero, self.one).mask(self.include_prior),
                )
                t = (
                    cell_time * t_scale + t + gene_offset
                    if self.t_scale_on and t_scale is not None
                    else cell_time + t
                )
            else:
                t = pyro.sample(
                    "latent_time",
                    LogNormal(self.zero, self.one).mask(self.include_prior),
                )

            state = (
                pyro.sample(
                    "cell_gene_state",
                    Bernoulli(logits=t - switching),
                    infer={"enumerate": "sequential"},
                )
                == self.zero
            )
            u0_vec = torch.where(state, u0, u_inf)
            s0_vec = torch.where(state, s0, s_inf)
            alpha_vec = torch.where(state, alpha, self.zero)

            tau = softplus(torch.where(state, t - gene_offset, t - switching))

            ut, st = mRNA(tau, u0_vec, s0_vec, alpha_vec, beta, gamma)

            if self.latent_factor == "linear":
                regressor_output = torch.einsum(
                    "abc,cd->ad", cell_code, cell_codebook.squeeze()
                )
                regressor_u = softplus(
                    regressor_output[..., : self.num_genes].squeeze()
                    + u_pcs_mean
                )
                regressor_s = softplus(
                    regressor_output[..., self.num_genes :].squeeze()
                    + s_pcs_mean
                )
            if self.latent_factor_operation == "selection":
                ut = torch.where(
                    cellgene_type == self.one,
                    ut * u_scale / s_scale,
                    softplus(regressor_u),
                )
                st = torch.where(
                    cellgene_type == self.one, st, softplus(regressor_s)
                )
            elif self.latent_factor_operation == "sum":
                ut = ut * u_scale / s_scale + regressor_u
                st = st + regressor_s
            else:
                ut = ut * u_scale / s_scale
                st = st
        return ut, st


class VelocityModelAuto(AuxCellVelocityModel):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_rna(
        self,
        u_scale: torch.Tensor,
        s_scale: torch.Tensor,
        alpha: torch.Tensor,
        beta: torch.Tensor,
        gamma: torch.Tensor,
        t: torch.Tensor,
        u0: torch.Tensor,
        s0: torch.Tensor,
        t0: torch.Tensor,
        switching: Optional[torch.Tensor] = None,
        u_inf: Optional[torch.Tensor] = None,
        s_inf: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Computes the unspliced (u) and spliced (s) RNA expression levels given the model parameters.

        Args:
            u_scale (torch.Tensor): Scaling factor for unspliced expression.
            s_scale (torch.Tensor): Scaling factor for spliced expression.
            alpha (torch.Tensor): Transcription rate.
            beta (torch.Tensor): Splicing rate.
            gamma (torch.Tensor): Degradation rate.
            t (torch.Tensor): Cell time.
            u0 (torch.Tensor): Unspliced RNA initial expression.
            s0 (torch.Tensor): Spliced RNA initial expression.
            t0 (torch.Tensor): Initial cell time.
            switching (Optional[torch.Tensor], optional): Switching time. Default is None.
            u_inf (Optional[torch.Tensor], optional): Unspliced RNA expression at switching time. Default is None.
            s_inf (Optional[torch.Tensor], optional): Spliced RNA expression at switching time. Default is None.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: The unspliced (u) and spliced (s) RNA expression levels.

        Examples:
            >>> import torch
            >>> from pyrovelocity._velocity_model_deprecated import VelocityModelAuto
            >>> model = VelocityModelAuto(
            ...             3,
            ...             4,
            ...             "Poisson",
            ...             True,
            ...             False,
            ...             2,
            ...             "none",
            ...             latent_factor_operation="selection",
            ...             latent_factor_size=10,
            ...             include_prior=False,
            ...             num_aux_cells=0,
            ...             only_cell_times=True,
            ...             decoder_on=False,
            ...             add_offset=False,
            ...             correct_library_size=True,
            ...             guide_type="auto_t0_constraint",
            ...             cell_specific_kinetics=None,
            ...             **{}
            ...         )
            >>> u, s = model.get_rna(
            ...            u_scale=torch.tensor([0.9793, 1.0567, 0.8610, 0.9304], device="cpu"),
            ...            s_scale=torch.tensor(1.0),
            ...            alpha=torch.tensor([0.4869, 1.5997, 1.3962, 0.5038], device="cpu"),
            ...            beta=torch.tensor([0.5403, 1.1192, 0.9912, 1.1783], device="cpu"),
            ...            gamma=torch.tensor([1.9612, 0.5533, 2.1050, 4.9345], device="cpu"),
            ...            t=torch.tensor([[0.4230], [0.5119], [0.2689]], device="cpu"),
            ...            u0=torch.tensor(0.0),
            ...            s0=torch.tensor(0.0),
            ...            t0=torch.tensor([-0.4867, 0.5581, -0.6957, 0.6028], device="cpu"),
            ...            switching=torch.tensor([1.1886, 1.1227, 0.6789, 4.1003], device="cpu"),
            ...            u_inf=torch.tensor([0.5367, 0.6695, 1.0479, 0.4206], device="cpu"),
            ...            s_inf=torch.tensor([0.1132, 0.2100, 0.3750, 0.0999], device="cpu"),
            >>>        )
        """
        if self.cell_specific_kinetics is None:
            if (
                self.guide_type != "auto"
                and version.parse(pyro.__version__) > version.parse("1.8.1")
                or self.guide_type == "auto"
            ):
                enum = "parallel"
            elif (
                self.guide_type != "auto"
                and not pyro.__version__.startswith("1.8.1")
                and pyro.__version__.startswith("1.6.0")
            ):
                enum = "sequential"

            state = (
                pyro.sample(
                    "cell_gene_state",
                    Bernoulli(logits=t - switching),
                    infer={"enumerate": enum},
                )
                == self.zero
            )
            alpha_off = self.zero
            u0_vec = torch.where(state, u0, u_inf)
            s0_vec = torch.where(state, s0, s_inf)
            alpha_vec = torch.where(state, alpha, alpha_off)
            tau = softplus(torch.where(state, t - t0, t - switching))
        else:
            u0_vec = u0
            s0_vec = s0
            alpha_vec = alpha
            tau = softplus(t - t0)

        ut, st = mRNA(tau, u0_vec, s0_vec, alpha_vec, beta, gamma)
        ut = ut * u_scale / s_scale
        return ut, st

    def forward(
        self,
        u_obs: Optional[torch.Tensor] = None,
        s_obs: Optional[torch.Tensor] = None,
        u_log_library: Optional[torch.Tensor] = None,
        s_log_library: Optional[torch.Tensor] = None,
        u_log_library_loc: Optional[torch.Tensor] = None,
        s_log_library_loc: Optional[torch.Tensor] = None,
        u_log_library_scale: Optional[torch.Tensor] = None,
        s_log_library_scale: Optional[torch.Tensor] = None,
        ind_x: Optional[torch.Tensor] = None,
        cell_state: Optional[torch.Tensor] = None,
        time_info: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Defines the forward model, which computes the unspliced (u) and spliced (s) RNA expression levels given the observations and model parameters.

        Args:
            u_obs (Optional[torch.Tensor], optional): Observed unspliced RNA expression. Default is None.
            s_obs (Optional[torch.Tensor], optional): Observed spliced RNA expression. Default is None.
            u_log_library (Optional[torch.Tensor], optional): Log-transformed library size for unspliced RNA. Default is None.
            s_log_library (Optional[torch.Tensor], optional): Log-transformed library size for spliced RNA. Default is None.
            u_log_library_loc (Optional[torch.Tensor], optional): Mean of log-transformed library size for unspliced RNA. Default is None.
            s_log_library_loc (Optional[torch.Tensor], optional): Mean of log-transformed library size for spliced RNA. Default is None.
            u_log_library_scale (Optional[torch.Tensor], optional): Scale of log-transformed library size for unspliced RNA. Default is None.
            s_log_library_scale (Optional[torch.Tensor], optional): Scale of log-transformed library size for spliced RNA. Default is None.
            ind_x (Optional[torch.Tensor], optional): Indices for the cells. Default is None.
            cell_state (Optional[torch.Tensor], optional): Cell state information. Default is None.
            time_info (Optional[torch.Tensor], optional): Time information for the cells. Default is None.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: The unspliced (u) and spliced (s) RNA expression levels.

        Examples:
            >>> import torch
            >>> from pyrovelocity._velocity_model_deprecated import VelocityModelAuto
            >>> u_obs=torch.tensor(
            ...     [[33.,  1.,  7.,  1.],
            ...     [12., 30., 11.,  3.],
            ...     [ 1.,  1.,  8.,  5.]],
            ...     device="cpu",
            >>> )
            >>> s_obs=torch.tensor(
            ...     [[32.0, 0.0, 6.0, 0.0],
            ...     [11.0, 29.0, 10.0, 2.0],
            ...     [0.0, 0.0, 7.0, 4.0]],
            ...     device="cpu",
            >>> )
            >>> u_log_library=torch.tensor([[3.7377], [4.0254], [2.7081]], device="cpu")
            >>> s_log_library=torch.tensor([[3.6376], [3.9512], [2.3979]], device="cpu")
            >>> u_log_library_loc=torch.tensor([[3.4904], [3.4904], [3.4904]], device="cpu")
            >>> s_log_library_loc=torch.tensor([[3.3289], [3.3289], [3.3289]], device="cpu")
            >>> u_log_library_scale=torch.tensor([[0.6926], [0.6926], [0.6926]], device="cpu")
            >>> s_log_library_scale=torch.tensor([[0.8214], [0.8214], [0.8214]], device="cpu")
            >>> ind_x=torch.tensor([2, 0, 1], device="cpu")
            >>> model = VelocityModelAuto(3,4)
            >>> u, s = model.forward(
            >>>            u_obs,
            >>>            s_obs,
            >>>            u_log_library,
            >>>            s_log_library,
            >>>            u_log_library_loc,
            >>>            s_log_library_loc,
            >>>            u_log_library_scale,
            >>>            s_log_library_scale,
            >>>            ind_x,
            >>>        )
            >>> u, s
            (tensor([[33.,  1.,  7.,  1.],
                    [12., 30., 11.,  3.],
                    [ 1.,  1.,  8.,  5.]]),
            tensor([[32.,  0.,  6.,  0.],
                    [11., 29., 10.,  2.],
                    [ 0.,  0.,  7.,  4.]]))
        """

        cell_plate, gene_plate = self.create_plates(
            u_obs,
            s_obs,
            u_log_library,
            s_log_library,
            u_log_library_loc,
            s_log_library_loc,
            u_log_library_scale,
            s_log_library_scale,
            ind_x,
            cell_state,
            time_info,
        )

        with gene_plate, poutine.mask(mask=self.include_prior):
            alpha = self.alpha
            gamma = self.gamma
            beta = self.beta

            if self.add_offset:
                u0 = pyro.sample("u_offset", LogNormal(self.zero, self.one))
                s0 = pyro.sample("s_offset", LogNormal(self.zero, self.one))
            else:
                s0 = u0 = self.zero

            t0 = pyro.sample("t0", Normal(self.zero, self.one))

            if (self.likelihood == "Normal") or (self.guide_type == "auto"):
                u_scale = self.u_scale
                s_scale = self.one
            else:
                u_scale = s_scale = self.one

            if self.likelihood == "Normal":
                s_scale = self.s_scale
            if self.cell_specific_kinetics is None:
                dt_switching = self.dt_switching
                u_inf, s_inf = mRNA(dt_switching, u0, s0, alpha, beta, gamma)
                u_inf = pyro.deterministic("u_inf", u_inf, event_dim=0)
                s_inf = pyro.deterministic("s_inf", s_inf, event_dim=0)
                switching = pyro.deterministic(
                    "switching", dt_switching + t0, event_dim=0
                )
            else:
                switching = u_inf = s_inf = None

        with cell_plate:
            t = pyro.sample(
                "cell_time",
                LogNormal(self.zero, self.one).mask(self.include_prior),
            )

        with cell_plate:
            u_cell_size_coef = ut_coef = s_cell_size_coef = st_coef = None
            u_read_depth = s_read_depth = None
            if self.correct_library_size and (self.likelihood != "Normal"):
                if self.guide_type == "velocity_auto":
                    u_read_depth = torch.exp(u_log_library)
                    s_read_depth = torch.exp(s_log_library)
                else:
                    u_read_depth = pyro.sample(
                        "u_read_depth",
                        LogNormal(u_log_library, u_log_library_scale),
                    )
                    s_read_depth = pyro.sample(
                        "s_read_depth",
                        LogNormal(s_log_library, s_log_library_scale),
                    )
                    if self.correct_library_size == "cell_size_regress":
                        u_cell_size_coef = pyro.sample(
                            "u_cell_size_coef", Normal(self.zero, self.one)
                        )
                        ut_coef = pyro.sample(
                            "ut_coef", Normal(self.zero, self.one)
                        )
                        s_cell_size_coef = pyro.sample(
                            "s_cell_size_coef", Normal(self.zero, self.one)
                        )
                        st_coef = pyro.sample(
                            "st_coef", Normal(self.zero, self.one)
                        )
            with gene_plate:
                if self.guide_type in [
                    "auto_t0_constraint",
                    "velocity_auto_t0_constraint",
                ]:
                    pyro.sample(
                        "time_constraint",
                        Bernoulli(logits=t - t0),
                        obs=self.one,
                    )
                ut, st = self.get_rna(
                    u_scale,
                    s_scale,
                    alpha,
                    beta,
                    gamma,
                    t,
                    u0,
                    s0,
                    t0,
                    switching,
                    u_inf,
                    s_inf,
                )
                u_dist, s_dist = self.get_likelihood(
                    ut,
                    st,
                    u_log_library,
                    s_log_library,
                    u_scale,
                    s_scale,
                    u_read_depth=u_read_depth,
                    s_read_depth=s_read_depth,
                    u_cell_size_coef=u_cell_size_coef,
                    ut_coef=ut_coef,
                    s_cell_size_coef=s_cell_size_coef,
                    st_coef=st_coef,
                )
                u = pyro.sample("u", u_dist, obs=u_obs)
                s = pyro.sample("s", s_dist, obs=s_obs)
        return u, s



import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scvelo as scv
import seaborn as sns
from matplotlib import ticker
from mpl_toolkits.axes_grid1 import make_axes_locatable


def rainbowplot(
    volcano_data,
    adata,
    posterior_samples,
    fig=None,
    genes=None,
    data=["st", "ut"],
    cell_state="clusters",
    basis="umap",
    num_genes=5,
    add_line=True,
    negative=False,
    scvelo_colors=False,
) -> None:
    set_font_size(7)

    if genes is None:
        genes = get_genes(volcano_data, num_genes, negative)
    number_of_genes = len(genes)

    subplot_height = 1  # Set the height of each subplot
    subplot_width = (
        subplot_height * 2.0 * 3
    )  # Width for each row of subplots to keep 1:1 aspect ratio

    if fig is None:
        fig, ax = plt.subplots(
            number_of_genes,
            3,
            figsize=(subplot_width, subplot_height * number_of_genes),
        )

    if scvelo_colors:
        colors = setup_scvelo_colors(adata, cell_state)
    else:
        colors = setup_colors(adata, cell_state)

    st, ut = get_posterior_samples(data, posterior_samples)

    for n, gene in enumerate(genes):
        ress = get_data(gene, st, ut, adata, cell_state, posterior_samples)
        ax1 = ax[n, 1]
        ax2 = ax[n, 0]
        ax3 = ax[n, 2]
        if n == 0:
            ax1.set_title("Rainbow plot", fontsize=7)
            ax2.set_title("Phase portrait", fontsize=7)
            ax3.set_title("Denoised spliced", fontsize=7)
        plot_gene(ax1, ress, colors, add_line)
        scatterplot(ax2, ress, colors)
        (index,) = np.where(adata.var_names == gene)
        im = ax3.scatter(
            adata.obsm[f"X_{basis}"][:, 0],
            adata.obsm[f"X_{basis}"][:, 1],
            s=3,
            c=st[:, index].flatten(),
            cmap="RdBu_r",
        )
        set_colorbar(im, ax3, labelsize=5, fig=fig, rainbow=True)
        ax3.axis("off")
        set_labels(ax1, ax2, ax3, gene, number_of_genes, ress, n)

    sns.despine()
    fig.subplots_adjust(
        hspace=0.5,
        wspace=0.5,  # left=0.15, right=0.4, top=0.92, bottom=0.12
    )
    return fig


def set_subfigure_titles(ax, n):
    if n == 0:
        ax[0].set_title("Rainbow plot", fontsize=7)
        ax[1].set_title("Phase portrait", fontsize=7)


def plot_gene_data(ax, ress, colors, add_line):
    plot_gene(ax[1], ress, colors, add_line)
    scatterplot(ax[0], ress, colors)


def scatter_gene(ax, adata, st, gene, basis):
    (index,) = np.where(adata.var_names == gene)
    return ax.scatter(
        adata.obsm[f"X_{basis}"][:, 0],
        adata.obsm[f"X_{basis}"][:, 1],
        s=3,
        c=st[:, index].flatten(),
        cmap="RdBu_r",
    )


def adjust_subfigure(subfig):
    subfig[0].subplots_adjust(
        hspace=0.8, wspace=1.4, left=0.32, right=0.94, top=0.92, bottom=0.12
    )
    subfig[1].subplots_adjust(
        hspace=0.8, wspace=0.4, left=0.2, right=0.7, top=0.92, bottom=0.08
    )

    subfig[0].text(
        -0.025,
        0.58,
        "unspliced expression",
        size=7,
        rotation="vertical",
        va="center",
    )
    subfig[0].text(
        0.552,
        0.58,
        "spliced expression",
        size=7,
        rotation="vertical",
        va="center",
    )

    sns.despine()


def pareto_frontier_genes(volcano_data, num_genes):
    volcano_data = volcano_data.loc[
        ~volcano_data.index.str.contains(("^Rpl|^Rps"), case=False)
    ]
    pareto_frontier = pd.DataFrame()

    if len(volcano_data) < num_genes:
        print(
            f"\nWarning: Not enough genes in the input data.\n"
            f"Only {len(volcano_data)} genes were found,\n"
            f"but {num_genes} were requested.\n"
            "Attempting to return as many as possible.\n"
        )

    while len(pareto_frontier) < num_genes and len(volcano_data) > 0:
        sorted_data = volcano_data.sort_values(
            by=["mean_mae", "time_correlation"], ascending=[False, False]
        )
        pareto_frontier_current = sorted_data.iloc[:1]
        for i in range(1, len(sorted_data)):
            if (
                sorted_data["time_correlation"].iloc[i]
                >= pareto_frontier_current["time_correlation"].iloc[-1]
            ):
                pareto_frontier_current = pareto_frontier_current.append(
                    sorted_data.iloc[i]
                )

        pareto_frontier = pareto_frontier.append(pareto_frontier_current)
        volcano_data = volcano_data.loc[
            ~volcano_data.index.isin(pareto_frontier.index)
        ]
        print(
            f"\nFound {len(pareto_frontier_current)} genes on the current Pareto frontier:\n\n"
            f"  {pareto_frontier_current.index.tolist()}\n\n"
            f"Total genes found: {len(pareto_frontier)}."
        )

    pareto_frontier = pareto_frontier.sort_values(
        by="time_correlation", ascending=False
    )

    if len(pareto_frontier) > num_genes:
        print(
            f"Found more than {num_genes} genes on the Pareto frontiers. "
            f"Returning the top {num_genes} genes."
        )
        pareto_frontier = pareto_frontier.head(num_genes)

    return pareto_frontier.index.tolist()


def set_font_size(size: int):
    matplotlib.rcParams.update({"font.size": size})


def get_genes(volcano_data, num_genes, negative):
    return (
        volcano_data.sort_values("mean_mae", ascending=False)
        .head(300)
        .sort_values("time_correlation", ascending=negative)
        .head(num_genes)
        .index
    )


def setup_scvelo_colors(adata, cell_state, basis):
    scv.pl.scatter(
        adata,
        basis=basis,
        fontsize=7,
        legend_loc="on data",
        legend_fontsize=7,
        color=cell_state,
        show=False,
    )
    return dict(
        zip(adata.obs.state_info.cat.categories, adata.uns["state_info_colors"])
    )


def setup_colors(adata, cell_state):
    clusters = adata.obs.loc[:, cell_state]
    return dict(
        zip(
            clusters.cat.categories,
            sns.color_palette("deep", clusters.cat.categories.shape[0]),
        )
    )


def get_posterior_samples(data, posterior_samples):
    if (data[0] in posterior_samples) and (data[1] in posterior_samples):
        st = posterior_samples[data[0]].mean(0).squeeze()
        ut = posterior_samples[data[1]].mean(0).squeeze()
    else:
        st = posterior_samples["st_mean"]
        ut = posterior_samples["ut_mean"]
    return st, ut


def get_data(gene, st, ut, adata, cell_state, posterior_samples):
    (index,) = np.where(adata.var_names == gene)
    pos_mean_time = posterior_samples["cell_time"].mean(0).flatten()
    ress = pd.DataFrame(
        {
            "cell_time": pos_mean_time / pos_mean_time.max(),
            "cell_type": adata.obs[cell_state].values,
            "spliced": st[:, index].flatten(),
            "unspliced": ut[:, index].flatten(),
        }
    )
    return ress.sort_values("cell_time")


def plot_gene(ax1, ress, colors, add_line):
    sns.scatterplot(
        x="cell_time",
        y="spliced",
        data=ress,
        alpha=0.1,
        linewidth=0,
        edgecolor="none",
        hue="cell_type",
        palette=colors,
        ax=ax1,
        marker="o",
        legend=False,
        s=5,
    )
    if add_line:
        for row in range(ress.shape[0]):
            ax1.vlines(
                x=ress.cell_time[row],
                ymin=0,
                ymax=ress.spliced[row],
                colors=colors[ress.cell_type[row]],
                alpha=0.1,
            )


def set_labels(ax1, ax2, ax3, gene, ngenes, ress, n):
    if n == 0:
        ax3.set_title("Denoised spliced", fontsize=7)
    if n == ngenes - 1:
        ax1.set_xlabel("shared time", fontsize=7)
        ax1.set_ylabel("expression", fontsize=7)
        ax2.set_xlabel("spliced", fontsize=7)
        ax2.set_ylabel("uspliced", fontsize=7)
    else:
        ax1.set_xlabel("")
        ax2.set_xlabel("")
        ax1.set_ylabel("")
        ax2.set_ylabel("")
    ax2.text(
        -0.5,
        0.5,
        gene,
        fontsize=8,
        weight="normal",
        rotation=0,
        va="center",
        ha="right",
        transform=ax2.transAxes,
    )

    x_ticks = [0, np.power(10, get_closest_pow_of_10(ress["cell_time"].max()))]
    y_ticks = [0, np.power(10, get_closest_pow_of_10(ress["spliced"].max()))]
    ax1.set_xticks(x_ticks)
    ax1.set_yticks(y_ticks)
    ax1.get_xaxis().set_major_formatter(
        ticker.FuncFormatter(lambda x, _: construct_log_string(x))
    )
    ax1.get_yaxis().set_major_formatter(
        ticker.FuncFormatter(lambda x, _: construct_log_string(x))
    )
    ax1.tick_params(axis="both", which="major", labelsize=7)

    x_ticks = [0, np.power(10, get_closest_pow_of_10(ress["spliced"].max()))]
    y_ticks = [0, np.power(10, get_closest_pow_of_10(ress["unspliced"].max()))]
    ax2.set_xticks(x_ticks)
    ax2.set_yticks(y_ticks)
    ax2.get_xaxis().set_major_formatter(
        ticker.FuncFormatter(lambda x, _: construct_log_string(x))
    )
    ax2.get_yaxis().set_major_formatter(
        ticker.FuncFormatter(lambda x, _: construct_log_string(x))
    )
    ax2.tick_params(axis="both", which="major", labelsize=7)


def get_closest_pow_of_10(value):
    return np.floor(np.log10(value))


def construct_log_string(x):
    str_val = f"$10^{{{int(np.log10(x))}}}$" if x > 0 else "0"
    return rf"{str_val}"


def scatterplot(ax2, ress, colors):
    sns.scatterplot(
        x="spliced",
        y="unspliced",
        data=ress,
        alpha=0.4,
        linewidth=0,
        edgecolor="none",
        hue="cell_type",
        palette=colors,
        ax=ax2,
        marker="o",
        legend=False,
        s=3,
    )


def set_colorbar(
    smp,
    ax,
    orientation="vertical",
    labelsize=None,
    fig=None,
    position="right",
    rainbow=False,
):
    from matplotlib.ticker import MaxNLocator
    from mpl_toolkits.axes_grid1.inset_locator import inset_axes

    if position == "right" and (not rainbow):
        cax = inset_axes(ax, width="2%", height="30%", loc=4, borderpad=0)
        cb = fig.colorbar(smp, orientation=orientation, cax=cax)
    else:
        divider = make_axes_locatable(ax)
        cax = divider.append_axes(position, size="8%", pad=0.08)
        cb = fig.colorbar(smp, cax=cax, orientation=orientation, shrink=0.4)

    cb.ax.tick_params(labelsize=labelsize)
    cb.set_alpha(1)
    cb.draw_all()
    cb.locator = MaxNLocator(nbins=2, integer=True)

    if position == "left":
        cb.ax.yaxis.set_ticks_position("left")
    cb.update_ticks()



"""Main module."""
import click

from . import __version__


@click.command()
@click.version_option(version=__version__)
def main():
    """Command line interface for pyrovelocity."""
    click.echo("pyrovelocity")



import logging
import os

from rich.console import Console
from rich.logging import RichHandler
from rich.theme import Theme


def configure_logging(logger_name: str = "pyrovelocity") -> logging.Logger:
    """
    Configures logging with rich handler and checks for valid log level from
    environment.

    Defaults to `INFO` if no valid log level is found.
    """
    console_theme = Theme(
        {
            "logging.level.info": "dim cyan",
            "logging.level.warning": "magenta",
            "logging.level.error": "bold red",
            "logging.level.debug": "green",
        }
    )
    console = Console(theme=console_theme)
    rich_handler = RichHandler(
        console=console,
        rich_tracebacks=True,
        show_time=True,
        show_level=True,
        show_path=False,
        markup=True,
        log_time_format="[%X]",
    )
    valid_log_levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    log_level = os.getenv("LOG_LEVEL", "INFO").upper()

    if log_level not in valid_log_levels:
        log_level = "INFO"

    logging.basicConfig(
        level=log_level,
        format="%(name)s %(message)s",
        datefmt="[%X]",
        handlers=[rich_handler],
    )
    return logging.getLogger(logger_name)



import inspect
import os
from pathlib import Path
from typing import Optional, Tuple
from urllib.parse import unquote

import anndata
import numpy as np
import requests
import scanpy as sc
import validators
from beartype import beartype

import pyrovelocity.datasets
from pyrovelocity.logging import configure_logging
from pyrovelocity.utils import (
    generate_sample_data,
    print_anndata,
    print_attributes,
)

logger = configure_logging(__name__)


@beartype
def download_dataset(
    data_set_name: str,
    data_external_path: str = "data/external",
    source: str = "pyrovelocity",
    data_url: Optional[str] = None,
) -> Path:
    """
    Downloads a dataset based on the specified parameters and returns the path
    to the downloaded data.

    Args:
        data_set_name (str): Name of the dataset to download.
        download_file_name (str): Name of the dataset to download.
        download_path_root (Path): Path for downloading the dataset. Default is 'data/external'.
        data_external_path (Path): Path where the downloaded data will be stored. Default is 'data/external'.
        source (str): The source type of the dataset. Default is 'pyrovelocity'.
        data_url (str): URL from where the dataset can be downloaded. Takes precedence over source.

    Returns:
        Path: The path to the downloaded dataset file.

    Examples:
        >>> tmp = getfixture('tmp_path')
        >>> simulated_dataset = download_dataset(
        ...   'simulated',
        ...   str(tmp) + '/data/external',
        ...   'simulate',
        ... ) # xdoctest: +SKIP
        >>> simulated_dataset = download_dataset(
        ...   'simulated_path',
        ...   tmp / Path('data/external'),
        ...   'simulate',
        ... ) # xdoctest: +SKIP
        >>> pancreas_dataset = download_dataset(
        ...   data_set_name='pancreas_direct',
        ...   data_external_path=str(tmp) + '/data/external',
        ...   data_url='https://github.com/theislab/scvelo_notebooks/raw/master/data/Pancreas/endocrinogenesis_day15.h5ad',
        ... ) # xdoctest: +SKIP
        >>> pancreas_dataset = download_dataset(
        ...   'pancreas',
        ...   str(tmp) + '/data/external',
        ...   'pyrovelocity',
        ... ) # xdoctest: +SKIP
        >>> no_dataset = download_dataset(
        ...   'no_dataset',
        ...   str(tmp) + '/data/external',
        ...   'pyrovelocity',
        ... ) # xdoctest: +SKIP, +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        AttributeError
    """
    data_path = Path(data_external_path) / f"{data_set_name}.h5ad"
    data_external_path = Path(data_external_path)

    logger.info(
        f"\n\nVerifying existence of path for downloaded data: {data_external_path}\n"
    )
    data_external_path.mkdir(parents=True, exist_ok=True)

    logger.info(
        f"\n\nVerifying {data_set_name} data:\n"
        f"  data will be stored in {data_path}\n"
    )

    if data_path.is_file() and os.access(str(data_path), os.R_OK):
        logger.info(f"{data_path} exists")
        return data_path
    else:
        logger.info(f"Attempting to download {data_set_name} data...")
        if data_url is not None:
            logger.info(f"Validating URL {data_url}...")
            is_valid_url, valid_url_message = validate_url_and_file(data_url)
            if not is_valid_url:
                raise ValueError(
                    f"Invalid URL: {data_url}\n{valid_url_message}\n"
                )
            else:
                logger.info(valid_url_message)
            try:
                adata = sc.read(str(data_path), backup_url=data_url, cache=True)
            except Exception as e:
                logger.error(f"Failed to download from URL {data_url}: {e}")
        elif source == "pyrovelocity":
            logger.info(
                f"Downloading {data_set_name} data with pyrovelocity..."
            )
            try:
                download_method = getattr(pyrovelocity.datasets, data_set_name)
            except AttributeError:
                available_datasets = [
                    name
                    for name, func in inspect.getmembers(
                        pyrovelocity.datasets, inspect.isfunction
                    )
                    if func.__module__ == pyrovelocity.datasets.__name__
                ]
                error_message = (
                    f"Dataset '{data_set_name}' not found in pyrovelocity.datasets. "
                    f"Available datasets are: {available_datasets}"
                    f"You can specify a URL that resolves to a .h5ad file with"
                    f"the `data_url` argument instead of a source."
                )
                logger.error(error_message)
                raise
            else:
                adata = download_method(file_path=data_path)
        elif source == "simulate":
            logger.info(f"Generating {data_set_name} data from simulation...")
            adata = generate_sample_data(
                n_obs=3000,
                n_vars=1000,
                noise_model="gillespie",
                random_seed=99,
            )
            adata.write(str(data_path))
        else:
            raise ValueError(
                f"Invalid source: {source}\n"
                f"or URL: {data_url}\n"
                f"Please specify a valid source or URL that resolves to a .h5ad file."
            )

        print_attributes(adata)
        print_anndata(adata)

        if data_path.is_file() and os.access(str(data_path), os.R_OK):
            logger.info(f"successfully downloaded {data_path}")
        else:
            logger.error(f"cannot find and read {data_path}")

        return data_path


@beartype
def validate_url_and_file(url: str) -> Tuple[bool, str]:
    """
    Validates a given URL format and checks if it leads to a .h5ad file larger
    than 1MB.

    This function first checks if the URL is valid using the
    validators.url method. Then, it makes a HEAD request to the URL to fetch
    headers without downloading the entire file. It checks the Content-Type and
    Content-Disposition headers for the .h5ad file extension and the Content-
    Length header to determine the file size.

    Args:
        url (str): The URL to validate and check the file type and size.

    Returns:
        tuple: A tuple containing a boolean and a validation message.
               The boolean is True if the URL is valid, leads to a .h5ad file, and the file is larger than 1MB.
               The message provides additional information about the validation.

    Raises:
        requests.RequestException: If an error occurs during the HEAD request.

    Examples:
        >>> is_valid, message = validate_url_and_file("https://storage.googleapis.com/pyrovelocity/data/pbmc5k.h5ad")
        >>> logger.info(f"valid: {is_valid}\n{message}")

        >>> is_valid, message = validate_url_and_file("http?/invalid.url/file.txt")
        >>> logger.info(f"valid: {is_valid}\n{message}")

        >>> is_valid, message = validate_url_and_file("https://invalid.url/file.txt")
        >>> logger.info(f"valid: {is_valid}\n{message}")
    """
    if not validators.url(url):
        return False, "Invalid URL format"

    try:
        response = requests.head(url, allow_redirects=True)
        response.raise_for_status()

        content_disposition = response.headers.get("Content-Disposition")
        if content_disposition:
            filename = content_disposition.split("filename=")[-1]
            filename = unquote(filename).strip('"')
        else:
            filename = url.split("/")[-1]

        if not filename.endswith(".h5ad"):
            return False, "The file does not have an .h5ad extension"

        content_length = response.headers.get("Content-Length")
        if content_length and int(content_length) <= 1_000_000:
            return False, "The file size is less than or equal to 1MB"
        else:
            content_length_mb = int(content_length) / 1_000_000

        return (
            True,
            f"URL validated and file is an .h5ad file named {filename} with size {content_length_mb:.1f} MB",
        )
    except requests.RequestException as e:
        return False, f"Error occurred: {e}"


@beartype
def subset(
    file_path: Optional[str | Path] = None,
    adata: Optional[anndata._core.anndata.AnnData] = None,
    n_obs: int = 100,
    save_subset: bool = False,
    output_path: Optional[str | Path] = None,
) -> Tuple[anndata._core.anndata.AnnData, str | Path | None]:
    """
    Randomly sample observations from a dataset given by file path or AnnData object.

    Args:
        file_path (str): Path to a .h5ad file containing a dataset. Takes precedence over adata.
        adata (AnnData): AnnData object. If None, file_path must be provided.
        n_obs (int): Number of observations to sample. Defaults to 100.
        save_subset (bool): If True, save the subset to a file. Defaults to False.
        output_path (str): Path to save the subset. Defaults to None.

    Raises:
        ValueError: If neither file_path nor adata is provided.

    Returns:
        AnnData: Subset of the dataset.

    Examples:
        >>> data_path = download_dataset(data_set_name="pancreas") # xdoctest: +SKIP
        >>> adata = subset(file_path=data_path, n_obs=100, save_subset=True) # xdoctest: +SKIP
        >>> print_anndata(adata) # xdoctest: +SKIP
        >>> print_attributes(adata) # xdoctest: +SKIP
    """
    if file_path is not None:
        file_path = Path(file_path)
        adata = sc.read(file_path, cache=True)
    if adata is None:
        raise ValueError("Either file_path or adata must be provided")

    if n_obs > adata.n_obs:
        logger.warning(
            f"n_obs ({n_obs}) is greater than the number of observations in the dataset ({adata.n_obs})"
        )
        n_obs = adata.n_obs

    selected_obs_indices = np.random.choice(adata.n_obs, n_obs)
    adata = adata[selected_obs_indices]
    adata.obs_names_make_unique()

    if save_subset:
        if output_path is None and file_path is not None:
            output_path = file_path.parent / Path(
                file_path.stem + f"_{n_obs}obs" + file_path.suffix
            )
        if output_path is None:
            raise ValueError(
                "output_path must be provided if save_subset is True and file_path is None"
            )
        adata.write(output_path)
        logger.info(f"saved {n_obs} obs subset: {output_path}")

    return adata.copy(), output_path



from pathlib import Path

import anndata
import scanpy as sc
import scvelo as scv
from beartype import beartype


@beartype
def pbmc5k(
    file_path: str | Path = "data/external/pbmc5k.h5ad",
) -> anndata._core.anndata.AnnData:
    """
    10x genomics 5k Peripheral blood mononuclear cells (PBMCs) from a healthy donor.

    https://www.10xgenomics.com/datasets/peripheral-blood-mononuclear-cells-pbm-cs-from-a-healthy-donor-chromium-connect-channel-1-3-1-standard-3-1-0

    Returns:
        Returns `AnnData` object
    """
    url = "https://storage.googleapis.com/pyrovelocity/data/pbmc5k.h5ad"
    adata = sc.read(file_path, backup_url=url, sparse=True, cache=True)
    return adata


@beartype
def pbmc10k(
    file_path: str | Path = "data/external/pbmc10k.h5ad",
) -> anndata._core.anndata.AnnData:
    """
    10x genomics 10k Human PBMCs.

    https://www.10xgenomics.com/datasets/10k-human-pbmcs-3-ht-v3-1-chromium-x-3-1-high

    Returns:
        Returns `AnnData` object
    """
    url = "https://storage.googleapis.com/pyrovelocity/data/pbmc10k.h5ad"
    adata = sc.read(file_path, backup_url=url, sparse=True, cache=True)
    return adata


@beartype
def pons(
    file_path: str | Path = "data/external/pons.h5ad",
) -> anndata._core.anndata.AnnData:
    """
    Pons oligodendrocyte data from `La Manno et al. (2018) <https://doi.org/https://doi.org/10.1038/s41586-018-0414-6>`__.

    Data originally obtained from https://pklab.med.harvard.edu/ruslan/velocity/oligos/oligos_info.rds
    and converted to AnnData format.

    Returns:
        Returns `AnnData` object
    """
    url = "https://storage.googleapis.com/pyrovelocity/data/oligo_lite.h5ad"
    adata = sc.read(file_path, backup_url=url, sparse=True, cache=True)
    return adata


@beartype
def larry(
    file_path: str | Path = "data/external/larry.h5ad",
) -> anndata._core.anndata.AnnData:
    """
    In vitro hematopoiesis LARRY dataset.

    Data from `Weinreb et al. (2020) <DOI: 10.1126/science.aaw3381>'
    https://figshare.com/ndownloader/articles/20780344/versions/1

    Returns:
        Returns `AnnData` object
    """
    url = "https://figshare.com/ndownloader/files/37028569"
    adata = sc.read(file_path, backup_url=url, sparse=True, cache=True)
    return adata


@beartype
def larry_neu(
    file_path: str | Path = "data/external/larry_neu.h5ad",
) -> anndata._core.anndata.AnnData:
    """
    In vitro hematopoiesis LARRY dataset.

    Subset of Data from `Weinreb et al. (2020) <DOI: 10.1126/science.aaw3381>'
    consisting of unipotent neutrophils.

    https://figshare.com/ndownloader/files/37028575

    Returns:
        Returns `AnnData` object
    """
    url = "https://figshare.com/ndownloader/files/37028575"
    adata = sc.read(file_path, backup_url=url, sparse=True, cache=True)
    adata = adata[adata.obs.state_info != "Centroid", :]
    return adata


@beartype
def larry_mono(
    file_path: str | Path = "data/external/larry_mono.h5ad",
) -> anndata._core.anndata.AnnData:
    """
    In vitro hematopoiesis LARRY dataset
    Subset of Data from `Weinreb et al. (2020) <DOI: 10.1126/science.aaw3381>'
    consisting of unipotent monocytes.

    https://figshare.com/ndownloader/files/37028572

    Returns:
        Returns `AnnData` object
    """
    url = "https://figshare.com/ndownloader/files/37028572"
    adata = sc.read(file_path, backup_url=url, sparse=True, cache=True)
    adata = adata[adata.obs.state_info != "Centroid", :]
    return adata


@beartype
def larry_cospar(
    file_path: str | Path = "data/external/larry_cospar.h5ad",
) -> anndata._core.anndata.AnnData:
    """
    In vitro hematopoiesis LARRY dataset with cospar-based fate estimates.

    Subset of Data from `Weinreb et al. (2020) <DOI: 10.1126/science.aaw3381>'

    Returns:
        Returns `AnnData` object
    """
    url = "https://storage.googleapis.com/pyrovelocity/data/larry_cospar.h5ad"
    adata = sc.read(file_path, backup_url=url, sparse=True, cache=True)
    return adata


@beartype
def larry_cytotrace(
    file_path: str | Path = "data/external/larry_cytotrace.h5ad",
) -> anndata._core.anndata.AnnData:
    """
    In vitro hematopoiesis LARRY dataset with cytotrace-based fate estimates.

    Subset of Data from `Weinreb et al. (2020) <DOI: 10.1126/science.aaw3381>'

    Returns:
        Returns `AnnData` object
    """
    url = (
        "https://storage.googleapis.com/pyrovelocity/data/larry_cytotrace.h5ad"
    )
    adata = sc.read(file_path, backup_url=url, sparse=True, cache=True)
    return adata


@beartype
def larry_dynamical(
    file_path: str | Path = "data/external/larry_dynamical.h5ad",
) -> anndata._core.anndata.AnnData:
    """
    The LARRY in vitro hematopoiesis dataset with scvelo dynamical model output.

    Subset of Data from `Weinreb et al. (2020) <DOI: 10.1126/science.aaw3381>'

    Returns:
        Returns `AnnData` object
    """
    url = (
        "https://storage.googleapis.com/pyrovelocity/data/larry_dynamical.h5ad"
    )
    adata = sc.read(file_path, backup_url=url, sparse=True, cache=True)
    return adata


@beartype
def larry_tips(
    file_path: str | Path = "data/external/larry_tips.h5ad",
) -> anndata._core.anndata.AnnData:
    """
    The differentiated subset of the LARRY in vitro hematopoiesis dataset.

    Subset of Data from `Weinreb et al. (2020) <DOI: 10.1126/science.aaw3381>'

    Returns:
        Returns `AnnData` object
    """
    adata = larry()
    adata = adata[adata.obs["time_info"] == 6.0]
    adata = adata[adata.obs["state_info"] != "Undifferentiated"]
    return adata


@beartype
def larry_multilineage(
    file_path: str | Path = "data/external/larry_multilineage.h5ad",
) -> anndata._core.anndata.AnnData:
    """
    The monocyte and neutrophil subset of the LARRY in vitro hematopoiesis dataset.

    Subset of Data from `Weinreb et al. (2020) <DOI: 10.1126/science.aaw3381>'

    Returns:
        Returns `AnnData` object
    """
    adata_larry_mono = larry_mono()
    adata_larry_neu = larry_neu()
    adata = adata_larry_mono.concatenate(adata_larry_neu)
    return adata


@beartype
def pancreas(
    file_path: str | Path = "data/external/pancreas.h5ad",
) -> anndata._core.anndata.AnnData:
    """
    Pancreatic endocrinogenesis data sourced from the scvelo library.

    Data from `Bastidas-Ponce et al. (2019) <https://doi.org/10.1242/dev.173849>`__
    will be downloaded from

    https://github.com/theislab/scvelo_notebooks/raw/master/data/Pancreas/endocrinogenesis_day15.h5ad

    Args:
        file_path (str, optional): Path to save file. Defaults to "data/external/pancreas.h5ad".

    Returns:
        Returns `AnnData` object
    """
    adata = scv.datasets.pancreas(file_path=file_path)
    return adata


@beartype
def bonemarrow(
    file_path: str | Path = "data/external/bonemarrow.h5ad",
) -> anndata._core.anndata.AnnData:
    """
    Human bone marrow data sourced from the scvelo library.

    Data from `Setty et al. (2019) <https://doi.org/10.1038/s41587-019-0068-4>`__
    will be downloaded from

    https://ndownloader.figshare.com/files/27686835

    Args:
        file_path (str, optional): Path to save file. Defaults to "data/external/bonemarrow.h5ad".

    Returns:
        Returns `AnnData` object
    """
    adata = scv.datasets.bonemarrow(file_path=file_path)
    return adata


@beartype
def pbmc68k(
    file_path: str | Path = "data/external/pbmc68k.h5ad"
) -> anndata._core.anndata.AnnData:
    """
    Peripheral blood mononuclear cells data sourced from the scvelo library.

    Data from `Zheng et al. (2017) <https://doi.org/10.1038/ncomms14049>`__
    will be downloaded from

    https://ndownloader.figshare.com/files/27686886

    Args:
        file_path (str, optional): Path to save file. Defaults to "data/external/pbmc68k.h5ad".

    Returns:
        Returns `AnnData` object
    """
    adata = scv.datasets.pbmc68k(file_path=file_path)
    scv.pp.remove_duplicate_cells(adata)
    adata.obsm["X_tsne"][:, 0] *= -1
    return adata



import dataclasses
import inspect
import sys
from dataclasses import field
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import get_type_hints

from mashumaro.mixins.json import DataClassJSONMixin
from sklearn.linear_model import LogisticRegression


def infer_type_from_default(value: Any) -> Type:
    """
    Infers or imputes a type from the default value of a parameter.
    Args:
        value: The default value of the parameter.
    Returns:
        The inferred type.
    """
    if value is None:
        return Optional[Any]
    elif value is inspect.Parameter.empty:
        return Any
    else:
        return type(value)


def create_dataclass_from_callable(
    callable_obj: Callable,
    overrides: Optional[Dict[str, Tuple[Type, Any]]] = None,
) -> List[Tuple[str, Type, Any]]:
    """
    Creates the fields of a dataclass from a `Callable` that includes all
    parameters of the callable as typed fields with default values inferred or
    taken from type hints. The function also accepts a dictionary containing
    parameter names together with a tuple of a type and default to allow
    specification of or override (un)typed defaults from the target callable.

    Args:
        callable_obj (Callable): The callable object to create a dataclass from.
        overrides (Optional[Dict[str, Tuple[Type, Any]]]): Dictionary to
        override inferred types and default values. Each dict value is a tuple
        (Type, default_value).

    Returns:
        Fields that can be used to construct a new dataclass type that
        represents the interface of the callable.

    Examples:
        >>> from pprint import pprint
        >>> custom_types_defaults: Dict[str, Tuple[Type, Any]] = {
        ...     "penalty": (str, "l2"),
        ...     "class_weight": (Optional[dict], None),
        ...     "random_state": (Optional[int], None),
        ...     "max_iter": (int, 2000),
        ...     "n_jobs": (Optional[int], None),
        ...     "l1_ratio": (Optional[float], None),
        ... }
        >>> fields = create_dataclass_from_callable(LogisticRegression, custom_types_defaults)
        >>> LogisticRegressionInterface = dataclasses.make_dataclass(
        ...     "LogisticRegressionInterface", fields, bases=(DataClassJSONMixin,)
        ... )
        >>> lr_instance = LogisticRegressionInterface()
        >>> isinstance(lr_instance, DataClassJSONMixin)
        True
        >>> pprint(lr_instance)
        LogisticRegressionInterface(penalty='l2',
                                    dual=False,
                                    tol=0.0001,
                                    C=1.0,
                                    fit_intercept=True,
                                    intercept_scaling=1,
                                    class_weight=None,
                                    random_state=None,
                                    solver='lbfgs',
                                    max_iter=2000,
                                    multi_class='auto',
                                    verbose=0,
                                    warm_start=False,
                                    n_jobs=None,
                                    l1_ratio=None)
    """
    if inspect.isclass(callable_obj):
        func = callable_obj.__init__
    else:
        func = callable_obj

    signature = inspect.signature(func)
    type_hints = get_type_hints(func)

    fields = []
    for name, param in signature.parameters.items():
        if name == "self":
            continue

        if overrides and name in overrides:
            field_type, default_value = overrides[name]
        else:
            inferred_type = infer_type_from_default(param.default)
            field_type = type_hints.get(name, inferred_type)
            default_value = (
                param.default
                if param.default is not inspect.Parameter.empty
                else dataclasses.field(default_factory=lambda: None)
            )

        fields.append((name, field_type, default_value))

    return fields


# -----------
# Deprecated
# -----------


class TypeInferenceError(Exception):
    pass


def infer_type_from_default_json(
    default: Any, name: str, custom_types: Optional[Dict[str, Type]] = None
) -> Type:
    """
    Infers the type from the default value of a parameter. Supports basic types
    like bool, int, float, str, list, and dict.

    Args:
        default: The default value of a parameter.
        name: The name of the parameter.
        custom_types: A dictionary of custom types for specific parameter names.

    Returns:
        The inferred type based on the default value, or Any if the type cannot be inferred.

    Raises:
        TypeInferenceError: If the type cannot be inferred and is not provided in custom_types.

    Examples:
        >>> infer_type_from_default_json(5, "example_int")
        <class 'int'>
        >>> infer_type_from_default_json(True, "example_bool")
        <class 'bool'>
        >>> infer_type_from_default_json("example", "example_str")
        <class 'str'>
        >>> infer_type_from_default_json(None, "example_none")  # Raises an exception
        Traceback (most recent call last):
        ...
        pyrovelocity.workflows.configuration.TypeInferenceError: Type for parameter 'example_none' with default value None cannot be inferred.
        Add this parameter to the custom_types dictionary.
    """

    if custom_types is None:
        custom_types = {}
    if name in custom_types:
        return custom_types[name]

    if default is None:
        type_inference_error_message = (
            f"Type for parameter '{name}' with default value None cannot be inferred.\n"
            "Add this parameter to the custom_types dictionary.\n"
        )
        raise TypeInferenceError(type_inference_error_message)

    if default is inspect.Parameter.empty:
        type_inference_error_message = (
            f"Type for parameter '{name}' with no default value cannot be inferred.\n"
            "Add this parameter to the custom_types dictionary.\n"
        )
        raise TypeInferenceError(type_inference_error_message)

    elif isinstance(default, bool):
        return bool
    elif isinstance(default, int):
        return int
    elif isinstance(default, float):
        return float
    elif isinstance(default, str):
        return str
    elif isinstance(default, list):
        return list
    elif isinstance(default, dict):
        return dict
    else:
        type_inference_error_message = (
            f"Type for parameter '{name}' with default value {default} cannot be inferred.\n"
            "Add this parameter to the custom_types dictionary.\n"
        )
        raise TypeInferenceError(type_inference_error_message)


def create_dataclass_from_callable_json(
    callable_obj: Callable, custom_types: Optional[Dict[str, Type]] = None
) -> Type:
    """
    Creates a dataclass from a callable object (such as a function or class
    constructor). This dataclass includes all parameters of the callable as
    fields, with types inferred or directly taken from type hints. Fields are
    assigned default values based on the callable's signature.

    Args:
        callable_obj: The callable object from which to create a dataclass.
        custom_types: An optional dictionary for overriding inferred types.

    Returns:
        A new dataclass type that represents the interface of the callable.

    Examples:
        >>> from dataclasses import dataclass
        >>> from dataclasses_json import dataclass_json
        ... # Example with sklearn's LogisticRegression
        >>> from sklearn.linear_model import LogisticRegression
        >>> logistic_regression_custom_types = {
        ...     "penalty": Optional[str],
        ...     "class_weight": Optional[dict],
        ...     "random_state": Optional[int],
        ...     "n_jobs": Optional[int],
        ...     "l1_ratio": Optional[float],
        ... }
        >>> BaseLogisticRegressionInterface = create_dataclass_from_callable_json(
        ...     LogisticRegression, logistic_regression_custom_types
        ... )
        >>> hasattr(BaseLogisticRegressionInterface, 'fit_intercept')
        True

        # Extending the base class with additional methods or properties
        >>> class LogisticRegressionInterface(BaseLogisticRegressionInterface):
        ...     def additional_method(self):
        ...         return "Some custom behavior"
        >>> LogisticRegressionInterface = dataclass_json(dataclass(LogisticRegressionInterface))
        >>> hasattr(LogisticRegressionInterface, 'additional_method')
        True

        # Using decorators without extending the base class
        >>> @dataclass_json
        ... @dataclass
        ... class LogisticRegressionInterface(create_dataclass_from_callable_json(
        ...     LogisticRegression, logistic_regression_custom_types)):
        ...     pass
        >>> hasattr(LogisticRegressionInterface, 'fit_intercept')
        True

        # Testing with a simple function and custom types
        >>> def example_func(a: int, b='hello', c: bool=False):
        ...     pass
        >>> example_func_custom_types = {"b": str}
        >>> ExampleFuncInterface = dataclass_json(
        ...     dataclass(create_dataclass_from_callable_json(
        ...     example_func, example_func_custom_types
        ...     ))
        ... )
        Traceback (most recent call last):
        ...
        pyrovelocity.workflows.configuration.TypeInferenceError: Type for parameter 'a' with no default value cannot be inferred.
        Add this parameter to the custom_types dictionary.
        >>> def example_func(a, b='hello', c: bool=False):
        ...     pass
        >>> example_func_custom_types = {"a": int, "b": str,}
        >>> ExampleFuncInterface = dataclass_json(
        ...     dataclass(create_dataclass_from_callable_json(
        ...     example_func, example_func_custom_types
        ...     ))
        ... )
        >>> example_instance = ExampleFuncInterface(a=5)
        >>> example_instance.b == 'hello'
        True
    """
    caller_module = sys._getframe(1).f_globals["__name__"]

    if inspect.isclass(callable_obj):
        func = callable_obj.__init__
    else:
        func = callable_obj

    signature = inspect.signature(func)
    type_hints = get_type_hints(func)

    class_attrs = {"__annotations__": {}}

    for name, param in signature.parameters.items():
        if name == "self":
            continue

        try:
            field_type = type_hints.get(
                name,
                infer_type_from_default_json(param.default, name, custom_types),
            )
        except TypeInferenceError as e:
            raise TypeInferenceError(str(e)) from e

        default = (
            param.default
            if param.default is not inspect.Parameter.empty
            else field(default_factory=lambda: None)
        )
        class_attrs[name] = (
            field(default=default) if default is field else default
        )
        class_attrs["__annotations__"][name] = field_type

    dataclass_name = f"{callable_obj.__name__}Interface"

    new_class = type(dataclass_name, (object,), class_attrs)
    new_class.__module__ = caller_module

    return new_class


# ---------------
# End Deprecated
# ---------------

if __name__ == "__main__":
    # Commented code here is primarily to support CLI or IDE debugger execution.
    # Otherwise, prefer to integrate tests and checks into the docstrings and
    # run pytest with `--xdoc` (default in this project).

    import pprint

    custom_types_defaults: Dict[str, Tuple[Type, Any]] = {
        # "penalty": (str, "l2"),
        # "dual": (bool, False),
        # "tol": (float, 1e-4),
        # "C": (float, 1.0),
        # "fit_intercept": (bool, True),
        # "intercept_scaling": (int, 1),
        # "class_weight": (Optional[dict], None),
        # "random_state": (Optional[int], None),
        # "solver": (str, "lbfgs"),
        "max_iter": (int, 2000),
        # "multi_class": (str, "auto"),
        # "verbose": (int, 0),
        # "warm_start": (bool, False),
        # "n_jobs": (Optional[int], None),
        # "l1_ratio": (Optional[float], None),
    }

    fields = create_dataclass_from_callable(
        LogisticRegression,
        custom_types_defaults,
        # {},
    )
    LogisticRegressionInterface = dataclasses.make_dataclass(
        "LogisticRegressionInterface", fields, bases=(DataClassJSONMixin,)
    )
    pprint.pprint(LogisticRegressionInterface())

    # from dataclasses import dataclass
    # from dataclasses_json import dataclass_json
    # from sklearn.linear_model import LogisticRegression
    # logistic_regression_custom_types = {
    #     "penalty": Optional[str],
    #     "class_weight": Optional[dict],
    #     "random_state": Optional[int],
    #     "n_jobs": Optional[int],
    #     "l1_ratio": Optional[float],
    # }
    # LogisticRegressionInterface = dataclass_json(
    #     dataclass(
    #         create_dataclass_from_callable_json(
    #             LogisticRegression, logistic_regression_custom_types
    #         )
    #     )
    # )
    # print("Annotations:", LogisticRegressionInterface.__annotations__)
    # print("Schema:", LogisticRegressionInterface().schema())



import logging
import os
import pickle
import sys
from statistics import harmonic_mean
from typing import Dict, Optional, Sequence, Union

import mlflow
import numpy as np
import pyro
import torch
from anndata import AnnData
from numpy import ndarray
from scvi.data import AnnDataManager
from scvi.data._constants import _SETUP_ARGS_KEY, _SETUP_METHOD_NAME
from scvi.data.fields import LayerField, NumericalObsField
from scvi.model._utils import parse_use_gpu_arg
from scvi.model.base import BaseModelClass
from scvi.model.base._utils import (
    _initialize_model,
    _load_saved_files,
    _validate_var_names,
)
from scvi.module.base import PyroBaseModuleClass

from pyrovelocity._trainer import VelocityTrainingMixin
from pyrovelocity._velocity_module import VelocityModule
from pyrovelocity.logging import configure_logging
from pyrovelocity.plot import (
    compute_mean_vector_field,
    compute_volcano_data,
    vector_field_uncertainty,
)
from pyrovelocity.utils import init_with_all_cells

logger = configure_logging(__name__)


class PyroVelocity(VelocityTrainingMixin, BaseModelClass):
    def __init__(
        self,
        adata: AnnData,
        input_type: str = "raw",
        shared_time: bool = True,
        model_type: str = "auto",
        guide_type: str = "auto",
        likelihood: str = "Poisson",
        t_scale_on: bool = False,
        plate_size: int = 2,
        latent_factor: str = "none",
        latent_factor_operation: str = "selection",
        inducing_point_size: int = 0,
        latent_factor_size: int = 0,
        include_prior: bool = False,
        use_gpu: Union[bool, int] = 0,
        init: bool = False,
        num_aux_cells: int = 0,
        only_cell_times: bool = True,
        decoder_on: bool = False,
        add_offset: bool = False,
        correct_library_size: Union[bool, str] = True,
        cell_specific_kinetics: Optional[str] = None,
        kinetics_num: Optional[int] = None,
    ) -> None:
        """
        PyroVelocity class for estimating RNA velocity and related tasks.

        Args:
            adata (AnnData): An AnnData object containing the gene expression data.
            input_type (str, optional): Type of input data. Can be "raw", "knn", or "raw_cpm". Defaults to "raw".
            shared_time (bool, optional): Whether to use shared time. Defaults to True.
            model_type (str, optional): Type of model to use. Defaults to "auto".
            guide_type (str, optional): Type of guide to use. Defaults to "auto".
            likelihood (str, optional): Type of likelihood to use. Defaults to "Poisson".
            t_scale_on (bool, optional): Whether to use t_scale. Defaults to False.
            plate_size (int, optional): Size of the plate. Defaults to 2.
            latent_factor (str, optional): Type of latent factor. Defaults to "none".
            latent_factor_operation (str, optional): Operation to perform on the latent factor. Defaults to "selection".
            inducing_point_size (int, optional): Size of inducing points. Defaults to 0.
            latent_factor_size (int, optional): Size of latent factors. Defaults to 0.
            include_prior (bool, optional): Whether to include prior information. Defaults to False.
            use_gpu (Union[bool, int], optional): Whether and which GPU to use. Defaults to 0. Can be False.
            init (bool, optional): Whether to initialize the model. Defaults to False.
            num_aux_cells (int, optional): Number of auxiliary cells. Defaults to 0.
            only_cell_times (bool, optional): Whether to use only cell times. Defaults to True.
            decoder_on (bool, optional): Whether to use decoder. Defaults to False.
            add_offset (bool, optional): Whether to add offset. Defaults to False.
            correct_library_size (Union[bool, str], optional): Whether to correct library size or method to correct. Defaults to True.
            cell_specific_kinetics (Optional[str], optional): Type of cell-specific kinetics. Defaults to None.
            kinetics_num (Optional[int], optional): Number of kinetics. Defaults to None.

        Examples:
            >>> import numpy as np
            >>> import anndata
            >>> from pyrovelocity.utils import pretty_log_dict, print_anndata, generate_sample_data
            >>> from pyrovelocity.preprocess import copy_raw_counts
            >>> from pyrovelocity._velocity import PyroVelocity
            >>> # setup sample data
            >>> n_obs = 10
            >>> n_vars = 5
            >>> adata = generate_sample_data(n_obs=n_obs, n_vars=n_vars)
            >>> copy_raw_counts(adata)
            >>> print_anndata(adata)
            >>> print(adata.X)
            >>> print(adata.layers['spliced'])
            >>> print(adata.layers['unspliced'])
            >>> print(adata.obs['u_lib_size_raw'])
            >>> print(adata.obs['s_lib_size_raw'])
            >>> PyroVelocity.setup_anndata(adata)
            >>> # train model
            >>> model = PyroVelocity(adata)
            >>> model.train(max_epochs=5, use_gpu=False)
            >>> posterior_samples = model.generate_posterior_samples(model.adata, num_samples=30)
            >>> print(posterior_samples.keys())
            >>> assert isinstance(posterior_samples, dict), f"Expected a dictionary, got {type(posterior_samples)}"
            >>> posterior_samples_log = pretty_log_dict(posterior_samples)
            >>> logger.debug(posterior_samples_log)
            >>> # print(posterior_samples_log)
            >>> model.save_model('save_pyrovelocity_doctest_model', overwrite=True)
            >>> model = PyroVelocity.load_model('save_pyrovelocity_doctest_model', adata, use_gpu=False)
            >>> # train model with
            >>> model = PyroVelocity(adata)
            >>> model.train_faster(max_epochs=5, use_gpu=False)
            >>> model.save_model('save_pyrovelocity_doctest_model', overwrite=True)
            >>> model = PyroVelocity.load_model('save_pyrovelocity_doctest_model', adata, use_gpu=False)
            >>> posterior_samples = model.generate_posterior_samples(model.adata, num_samples=30)
            >>> posterior_samples_log = pretty_log_dict(posterior_samples)
            >>> logger.debug(posterior_samples_log)
            >>> # print(posterior_samples_log)
            >>> print(posterior_samples.keys())
            >>> # train model with
            >>> model = PyroVelocity(adata)
            >>> model.train_faster_with_batch(batch_size=24, max_epochs=5, use_gpu=False)
            >>> model.save_model('save_pyrovelocity_doctest_model', overwrite=True)
            >>> model = PyroVelocity.load_model('save_pyrovelocity_doctest_model', adata, use_gpu=False)
            >>> posterior_samples = model.generate_posterior_samples(model.adata, num_samples=30)
            >>> posterior_samples_log = pretty_log_dict(posterior_samples)
            >>> logger.debug(posterior_samples_log)
            >>> # print(posterior_samples_log)
            >>> print(posterior_samples.keys())
        """
        self.use_gpu = use_gpu
        self.cell_specific_kinetics = cell_specific_kinetics
        self.k = kinetics_num
        if input_type == "knn":
            layers = ["Mu", "Ms"]
            assert likelihood in {"Normal", "LogNormal"}
            assert "Mu" in adata.layers
        elif input_type == "raw_cpm":
            layers = ["unspliced", "spliced"]
            assert likelihood in {"Normal", "LogNormal"}
        else:
            layers = ["raw_unspliced", "raw_spliced"]
            assert likelihood != "Normal"

        self.layers = layers
        self.input_type = input_type

        super().__init__(adata)
        if init:
            initial_values = init_with_all_cells(
                self.adata,
                input_type,
                shared_time,
                latent_factor,
                latent_factor_size,
                plate_size,
                num_aux_cells=num_aux_cells,
            )
        else:
            initial_values = {}
        logger.info(self.summary_stats)
        self.module = VelocityModule(
            self.summary_stats["n_cells"],
            self.summary_stats["n_vars"],
            model_type=model_type,
            guide_type=guide_type,
            likelihood=likelihood,
            shared_time=shared_time,
            t_scale_on=t_scale_on,
            plate_size=plate_size,
            latent_factor=latent_factor,
            latent_factor_operation=latent_factor_operation,
            latent_factor_size=latent_factor_size,
            inducing_point_size=inducing_point_size,
            include_prior=include_prior,
            use_gpu=use_gpu,
            num_aux_cells=num_aux_cells,
            only_cell_times=only_cell_times,
            decoder_on=decoder_on,
            add_offset=add_offset,
            correct_library_size=correct_library_size,
            cell_specific_kinetics=cell_specific_kinetics,
            kinetics_num=self.k,
            **initial_values,
        )
        self.num_cells = self.module.num_cells
        self._model_summary_string = """
        RNA velocity Pyro model with parameters:
        """
        self.init_params_ = self._get_init_params(locals())
        logger.info("The model has been initialized")

    def train(self, **kwargs):
        """
        Trains the PyroVelocity model using the provided data and configuration.

        The method leverages the Pyro library to train the model using the underlying
        data. It relies on the `VelocityTrainingMixin` to define the training logic.

        Args:

            **kwargs : dict, optional
                Additional keyword arguments to be passed to the underlying train method
                provided by the `VelocityTrainingMixin`.
        """
        pyro.enable_validation(True)
        super().train(**kwargs)

    def enum_parallel_predict(self):
        """work for parallel enumeration"""
        return

    @classmethod
    def setup_anndata(cls, adata: AnnData, *args, **kwargs):
        """Latest scvi-tools interface"""
        setup_method_args = cls._get_setup_method_args(**locals())

        adata.obs["u_lib_size"] = np.log(adata.obs["u_lib_size_raw"] + 1e-6)
        adata.obs["s_lib_size"] = np.log(adata.obs["s_lib_size_raw"] + 1e-6)

        adata.obs["u_lib_size_mean"] = adata.obs["u_lib_size"].mean()
        adata.obs["s_lib_size_mean"] = adata.obs["s_lib_size"].mean()
        adata.obs["u_lib_size_scale"] = adata.obs["u_lib_size"].std()
        adata.obs["s_lib_size_scale"] = adata.obs["s_lib_size"].std()
        adata.obs["ind_x"] = np.arange(adata.n_obs).astype("int64")

        anndata_fields = [
            LayerField("U", "raw_unspliced", is_count_data=True),
            LayerField("X", "raw_spliced", is_count_data=True),
            NumericalObsField("u_lib_size", "u_lib_size"),
            NumericalObsField("s_lib_size", "s_lib_size"),
            NumericalObsField("u_lib_size_mean", "u_lib_size_mean"),
            NumericalObsField("s_lib_size_mean", "s_lib_size_mean"),
            NumericalObsField("u_lib_size_scale", "u_lib_size_scale"),
            NumericalObsField("s_lib_size_scale", "s_lib_size_scale"),
            NumericalObsField("ind_x", "ind_x"),
        ]
        adata_manager = AnnDataManager(
            fields=anndata_fields, setup_method_args=setup_method_args
        )
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)

    def generate_posterior_samples(
        self,
        adata: Optional[AnnData] = None,
        indices: Optional[Sequence[int]] = None,
        batch_size: Optional[int] = None,
        num_samples: int = 100,
    ) -> Dict[str, ndarray]:
        """
        If the guide uses sequential enumeration, computes the posterior samples for
        the given data using the trained PyroVelocity model.

        The method generates posterior samples by running the trained model on the
        provided data and returns a dictionary containing samples for each parameter.

        Args:
            adata (AnnData, optional): Anndata object containing the data for which posterior samples
                are to be computed. If not provided, the anndata used to initialize the model will be used.
            indices (Sequence[int], optional): Indices of cells in `adata` for which the posterior
                samples are to be computed.
            batch_size (int, optional): The size of the mini-batches used during computation.
                If not provided, the entire dataset will be used.
            num_samples (int, default: 100): The number of posterior samples to compute for each parameter.

        Returns:
            Dict[str, ndarray]: A dictionary containing the posterior samples for each parameter.
        """
        self.module.eval()
        predictive = self.module.create_predictive(
            model=pyro.poutine.uncondition(self.module.model),
            num_samples=num_samples,
        )

        scdl = self._make_data_loader(
            adata=adata, indices=indices, batch_size=batch_size
        )

        with torch.no_grad(), pyro.poutine.mask(mask=False):
            posterior_samples = []
            for tensor in scdl:
                args, kwargs = self.module._get_fn_args_from_batch(tensor)
                posterior_sample = {
                    k: v.cpu().numpy()
                    for k, v in predictive(*args, **kwargs).items()
                }
                posterior_samples.append(posterior_sample)
            samples = {}
            for k in posterior_samples[0].keys():
                if k in [
                    "ut_norm",
                    "st_norm",
                    # "cell_gene_state",
                    "time_constraint",  # model 1 time constraint
                ]:  # skip unused variables
                    continue

                if "aux" in k:
                    samples[k] = posterior_samples[0][k]
                elif posterior_samples[0][k].shape[-2] == 1:
                    samples[k] = posterior_samples[0][k]
                    if k == "kinetics_prob":
                        samples[k] = np.concatenate(
                            [
                                posterior_samples[j][k]
                                for j in range(len(posterior_samples))
                            ],
                            axis=-3,
                        )
                else:
                    samples[k] = np.concatenate(
                        [
                            posterior_samples[j][k]
                            for j in range(len(posterior_samples))
                        ],
                        axis=-2,
                    )

                print(k, "before", sys.getsizeof(samples[k]))
        self.num_samples = num_samples
        return samples

    def get_mlflow_logs(self):
        return

    def compute_statistics_from_posterior_samples(
        self,
        adata,
        posterior_samples,
        vector_field_basis,
        ncpus_use,
    ):
        """reduce posterior samples by precomputing metrics."""
        if ("u_scale" in posterior_samples) and (
            "s_scale" in posterior_samples
        ):
            scale = posterior_samples["u_scale"] / posterior_samples["s_scale"]
        elif ("u_scale" in posterior_samples) and not (
            "s_scale" in posterior_samples
        ):
            scale = posterior_samples["u_scale"]
        else:
            scale = 1
        original_spaces_velocity_samples = (
            posterior_samples["beta"] * posterior_samples["ut"] / scale
            - posterior_samples["gamma"] * posterior_samples["st"]
        )
        original_spaces_embeds_magnitude = np.sqrt(
            (original_spaces_velocity_samples**2).sum(axis=-1)
        )

        (
            vector_field_posterior_samples,
            embeds_radian,
            fdri,
        ) = vector_field_uncertainty(
            adata,
            posterior_samples,
            basis=vector_field_basis,
            n_jobs=ncpus_use,
        )
        embeds_magnitude = np.sqrt(
            (vector_field_posterior_samples**2).sum(axis=-1)
        )

        mlflow.log_metric(
            "FDR_sig_frac", round((fdri < 0.05).sum() / fdri.shape[0], 3)
        )
        mlflow.log_metric("FDR_HMP", harmonic_mean(fdri))

        compute_mean_vector_field(
            posterior_samples=posterior_samples,
            adata=adata,
            basis=vector_field_basis,
            n_jobs=ncpus_use,
        )

        vector_field_posterior_mean = adata.obsm[
            f"velocity_pyro_{vector_field_basis}"
        ]

        gene_ranking, genes = compute_volcano_data(
            [posterior_samples], [adata], time_correlation_with="st"
        )
        gene_ranking = (
            gene_ranking.sort_values("mean_mae", ascending=False)
            .head(300)
            .sort_values("time_correlation", ascending=False)
        )
        posterior_samples["gene_ranking"] = gene_ranking
        posterior_samples[
            "original_spaces_embeds_magnitude"
        ] = original_spaces_embeds_magnitude
        posterior_samples["genes"] = genes
        posterior_samples[
            "vector_field_posterior_samples"
        ] = vector_field_posterior_samples
        posterior_samples[
            "vector_field_posterior_mean"
        ] = vector_field_posterior_mean
        posterior_samples["fdri"] = fdri
        posterior_samples["embeds_magnitude"] = embeds_magnitude
        print(embeds_radian.shape)
        posterior_samples["embeds_angle"] = embeds_radian
        posterior_samples["ut_mean"] = posterior_samples["ut"].mean(0).squeeze()
        posterior_samples["st_mean"] = posterior_samples["st"].mean(0).squeeze()

        (
            pca_vector_field_posterior_samples,
            pca_embeds_radian,
            pca_fdri,
        ) = vector_field_uncertainty(
            adata,
            posterior_samples,
            basis="pca",
            n_jobs=ncpus_use,
        )
        posterior_samples[
            "pca_vector_field_posterior_samples"
        ] = pca_vector_field_posterior_samples
        posterior_samples["pca_embeds_angle"] = pca_embeds_radian
        posterior_samples["pca_fdri"] = pca_fdri

        del posterior_samples["u"]
        del posterior_samples["s"]
        del posterior_samples["ut"]
        del posterior_samples["st"]
        return posterior_samples

    def save_pyrovelocity_data(self, posterior_samples, pyrovelocity_data_path):
        with open(pyrovelocity_data_path, "wb") as f:
            pickle.dump(posterior_samples, f)
        for k in posterior_samples:
            print(k, "after", sys.getsizeof(posterior_samples[k]))

    def save_model(
        self,
        dir_path: str,
        prefix: Optional[str] = None,
        overwrite: bool = True,
        save_anndata: bool = False,
        **anndata_write_kwargs,
    ) -> None:
        super().save(
            dir_path, prefix, overwrite, save_anndata, **anndata_write_kwargs
        )
        pyro.get_param_store().save(
            os.path.join(dir_path, "param_store_test.pt")
        )

    @classmethod
    def load_model(
        cls,
        dir_path: str,
        adata: Optional[AnnData] = None,
        use_gpu: Optional[Union[str, int, bool]] = None,
        prefix: Optional[str] = None,
        backup_url: Optional[str] = None,
    ):
        load_adata = adata is None
        _, _, device = parse_use_gpu_arg(use_gpu)

        (
            attr_dict,
            var_names,
            model_state_dict,
            new_adata,
        ) = _load_saved_files(
            dir_path,
            load_adata,
            map_location=device,
            prefix=prefix,
            backup_url=backup_url,
        )

        adata = new_adata if new_adata is not None else adata

        _validate_var_names(adata, var_names)

        registry = attr_dict.pop("registry_")
        method_name = registry.get(_SETUP_METHOD_NAME, "setup_anndata")
        getattr(cls, method_name)(
            adata, source_registry=registry, **registry[_SETUP_ARGS_KEY]
        )

        model = _initialize_model(cls, adata, attr_dict)
        print("---------initialize-------")

        for attr, val in attr_dict.items():
            setattr(model, attr, val)
        print("setattr")

        pyro.clear_param_store()
        old_history = model.history_
        try:
            model.module.load_state_dict(model_state_dict)
        except RuntimeError as err:
            if not isinstance(model.module, PyroBaseModuleClass):
                raise err
            logger.info("Preparing underlying module for load")
            try:
                print("train 1---")
                model.train(max_epochs=1, max_steps=1, use_gpu=use_gpu)
            except Exception:
                model.train(
                    max_epochs=1,
                    max_steps=1,
                    use_gpu=use_gpu,
                    batch_size=adata.shape[0],
                )
            model.module.load_state_dict(model_state_dict)

        model.history_ = old_history
        print("load finished.")
        model.to_device(device)
        model.module.eval()
        model._validate_anndata(adata)
        pyro.get_param_store().load(
            os.path.join(dir_path, "param_store_test.pt"), map_location=device
        )
        return model



"""Model training API for Pyro-Velocity."""
from typing import Dict, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
from anndata._core.anndata import AnnData
from numpy import ndarray
from pyro import poutine
from pyro.infer.autoguide import AutoNormal
from pyro.infer.autoguide.guides import AutoGuideList
from sklearn.model_selection import train_test_split

from pyrovelocity._velocity import PyroVelocity


def train_model(
    adata: AnnData,
    guide_type: str = "auto",
    model_type: str = "auto",
    svi_train: bool = False,  # svi_train alreadys turn off
    batch_size: int = -1,
    train_size: float = 1.0,
    use_gpu: int = 0,
    likelihood: str = "Poisson",
    num_samples: int = 30,
    log_every: int = 100,
    cell_state: str = "clusters",
    patient_improve: float = 5e-4,
    patient_init: int = 30,
    seed: int = 99,
    lr: float = 0.01,
    max_epochs: int = 3000,
    include_prior: bool = True,
    library_size: bool = True,
    offset: bool = False,
    input_type: str = "raw",
    cell_specific_kinetics: Optional[str] = None,
    kinetics_num: int = 2,
    loss_plot_path: str = "loss_plot.png",
) -> Tuple[PyroVelocity, Dict[str, ndarray]]:
    """
    Train a PyroVelocity model to provide probabilistic estimates of RNA velocity
    for single-cell RNA sequencing data with quantified splice variants.

    Args:
        adata (AnnData): An AnnData object containing the input data.
        guide_type (str, optional): The type of guide function for the Pyro model. Default is "auto".
        model_type (str, optional): The type of Pyro model. Default is "auto".
        svi_train (bool, optional): Whether to use Stochastic Variational Inference for training. Default is False.
        batch_size (int, optional): Batch size for training. Default is -1, which indicates using the full dataset.
        train_size (float, optional): Proportion of data to be used for training. Default is 1.0.
        use_gpu (int, optional): Whether to use GPU for training. Default is 0, which indicates not using GPU.
        likelihood (str, optional): Likelihood function for the Pyro model. Default is "Poisson".
        num_samples (int, optional): Number of posterior samples. Default is 30.
        log_every (int, optional): Frequency of logging progress. Default is 100.
        cell_state (str, optional): Cell state attribute in the AnnData object. Default is "clusters".
        patient_improve (float, optional): Minimum improvement in training loss for early stopping. Default is 5e-4.
        patient_init (int, optional): Number of initial training epochs before early stopping is enabled. Default is 30.
        seed (int, optional): Random seed for reproducibility. Default is 99.
        lr (float, optional): Learning rate for the optimizer. Default is 0.01.
        max_epochs (int, optional): Maximum number of training epochs. Default is 3000.
        include_prior (bool, optional): Whether to include prior information in the model. Default is True.
        library_size (bool, optional): Whether to correct for library size. Default is True.
        offset (bool, optional): Whether to add an offset to the model. Default is False.
        input_type (str, optional): Type of input data. Default is "raw".
        cell_specific_kinetics (Optional[str], optional): Name of the attribute containing cell-specific kinetics information. Default is None.
        kinetics_num (int, optional): Number of kinetics parameters. Default is 2.
        loss_plot_path (str, optional): Path to save the loss plot. Default is "loss_plot.png".

    Returns:
        Tuple[PyroVelocity, Dict[str, ndarray]]: A tuple containing the trained PyroVelocity model and a dictionary of posterior samples.

    Examples:
        >>> from pyrovelocity.api import train_model
        >>> from pyrovelocity.utils import generate_sample_data
        >>> from pyrovelocity.preprocess import copy_raw_counts
        >>> adata = generate_sample_data(random_seed=99)
        >>> copy_raw_counts(adata)
        >>> model, posterior_samples = train_model(adata, use_gpu=False, seed=99, max_epochs=200, loss_plot_path="loss_plot_docs.png")
    """
    PyroVelocity.setup_anndata(adata)

    model = PyroVelocity(
        adata,
        likelihood=likelihood,
        model_type=model_type,
        guide_type=guide_type,
        correct_library_size=library_size,
        add_offset=offset,
        include_prior=include_prior,
        input_type=input_type,
        cell_specific_kinetics=cell_specific_kinetics,
        kinetics_num=kinetics_num,
    )
    if svi_train and guide_type in {
        "velocity_auto",
        "velocity_auto_t0_constraint",
    }:
        if batch_size == -1:
            batch_size = adata.shape[0]
        model.train(
            max_epochs=max_epochs,
            lr=lr,
            use_gpu=use_gpu,
            batch_size=batch_size,
            train_size=train_size,
            valid_size=1 - train_size,
            check_val_every_n_epoch=1,
            early_stopping=True,
            patience=patient_init,
            min_delta=patient_improve,
        )

        fig, ax = plt.subplots()
        fig.set_size_inches(2.5, 1.5)
        ax.scatter(
            model.history_["elbo_train"].index[:-1],
            -model.history_["elbo_train"][:-1],
            label="Train",
        )
        if train_size < 1:
            ax.scatter(
                model.history_["elbo_validation"].index[:-1],
                -model.history_["elbo_validation"][:-1],
                label="Valid",
            )
        set_loss_plot_axes(ax)
        fig.savefig(loss_plot_path, facecolor="white", bbox_inches="tight")
        posterior_samples = model.generate_posterior_samples(
            model.adata, num_samples=num_samples, batch_size=512
        )
        return model, posterior_samples
    else:
        if train_size >= 1:  ##support velocity_auto_depth
            if batch_size == -1:
                batch_size = adata.shape[0]

            if batch_size >= adata.shape[0]:
                losses = model.train_faster(
                    max_epochs=max_epochs,
                    lr=lr,
                    use_gpu=use_gpu,
                    seed=seed,
                    patient_improve=patient_improve,
                    patient_init=patient_init,
                    log_every=log_every,
                )
            else:
                losses = model.train_faster_with_batch(
                    max_epochs=max_epochs,
                    batch_size=batch_size,
                    log_every=log_every,
                    lr=lr,
                    use_gpu=use_gpu,
                    seed=seed,
                    patient_improve=patient_improve,
                    patient_init=patient_init,
                )
            fig, ax = plt.subplots()
            fig.set_size_inches(2.5, 1.5)
            ax.scatter(
                np.arange(len(losses)),
                -np.array(losses),
                label="train",
                alpha=0.25,
            )
            set_loss_plot_axes(ax)
            posterior_samples = model.generate_posterior_samples(
                model.adata, num_samples=num_samples, batch_size=512
            )

            fig.savefig(loss_plot_path, facecolor="white", bbox_inches="tight")

            return model, posterior_samples
        else:  # train validation procedure
            if (
                guide_type == "velocity_auto_depth"
            ):  # velocity_auto, not supported (velocity_auto_depth, fails with error)
                raise

            indices = np.arange(adata.shape[0])
            train_ind, test_ind, cluster_train, cluster_test = train_test_split(
                indices,
                adata.obs.loc[:, cell_state].values,
                test_size=1 - train_size,
                random_state=seed,
                shuffle=False,
            )
            train_batch_size = (
                train_ind.shape[0] if batch_size == -1 else batch_size
            )
            losses = model.train_faster_with_batch(
                max_epochs=max_epochs,
                batch_size=train_batch_size,
                indices=train_ind,
                log_every=log_every,
                lr=lr,
                use_gpu=use_gpu,
                seed=seed,
                patient_improve=patient_improve,
                patient_init=patient_init,
            )
            posterior_samples = model.generate_posterior_samples(
                model.adata,
                num_samples=num_samples,
                indices=train_ind,
                batch_size=512,
            )

            test_batch_size = (
                test_ind.shape[0] if batch_size == -1 else batch_size
            )
            if guide_type in {"auto", "auto_t0_constraint"}:
                new_guide = AutoGuideList(
                    model.module._model,
                    create_plates=model.module._model.create_plates,
                )
                new_guide.append(
                    AutoNormal(
                        poutine.block(
                            model.module._model,
                            expose=[
                                "cell_time",
                                "u_read_depth",
                                "s_read_depth",
                            ],
                        ),
                        init_scale=0.1,
                    )
                )
                new_guide.append(
                    poutine.block(model.module._guide[-1], hide_types=["param"])
                )
                losses_test = model.train_faster_with_batch(
                    max_epochs=max_epochs,
                    batch_size=test_batch_size,
                    indices=test_ind,
                    new_valid_guide=new_guide,
                    log_every=log_every,
                    lr=lr,
                    use_gpu=use_gpu,
                    seed=seed,
                    elbo_name="-ELBO validation",
                )
            elif (
                guide_type
                in {
                    "velocity_auto",
                    "velocity_auto_t0_constraint",
                }
            ):  # velocity_auto, not supported (velocity_auto_depth fails with error)
                print("valid new guide")

                losses_test = model.train_faster_with_batch(
                    max_epochs=max_epochs,
                    batch_size=test_batch_size,
                    indices=test_ind,
                    log_every=log_every,
                    lr=lr,
                    use_gpu=use_gpu,
                    seed=seed,
                )
            else:
                raise
            pos_test = model.generate_posterior_samples(
                model.adata, num_samples=30, indices=test_ind, batch_size=512
            )
            fig, ax = plt.subplots()
            fig.set_size_inches(2.5, 1.5)
            ax.scatter(
                np.arange(len(losses)),
                -np.array(losses),
                label="train",
                alpha=0.25,
            )
            ax.scatter(
                np.arange(len(losses_test)),
                -np.array(losses_test),
                label="validation",
                alpha=0.25,
            )
            set_loss_plot_axes(ax)
        plt.legend()
        plt.savefig(loss_plot_path, facecolor="white", bbox_inches="tight")
        return posterior_samples, pos_test, train_ind, test_ind


def set_loss_plot_axes(ax):
    # ax.set_yscale('log')
    ax.set_yscale("symlog")
    ax.set_xlabel("Epochs")
    ax.set_ylabel("-ELBO")
